Package: treeshap
Title: Compute SHAP Values for Your Tree-Based Models Using the
        'TreeSHAP' Algorithm
Version: 0.4.0
Authors@R: c(
    person("Michael", "Mayer", , "mayermichael79@gmail.com", role = c("aut", "cre")),
    person("Konrad", "Komisarczyk", email = "komisarczykkonrad@gmail.com", role = "aut"),
    person("Pawel", "Kozminski", email = "pkozminski99@gmail.com", role = "aut"),
    person("Szymon", "Maksymiuk", role = "aut", comment = c(ORCID = "0000-0002-3120-1601")),
    person("Lorenz A.", "Kapsner", role = "ctb", comment = c(ORCID = "0000-0003-1866-860X")),
    person("Mikolaj", "Spytek", role = "ctb", comment = c(ORCID = "0000-0001-7111-2286")),
    person("Mateusz", "Krzyzinski", email = "krzyzinskimateusz23@gmail.com", role = "ctb", comment = c(ORCID = "0000-0001-6143-488X")),
    person("Przemyslaw", "Biecek", email = "przemyslaw.biecek@gmail.com", role = c("aut", "cph"), comment = c(ORCID = "0000-0001-8423-1823"))
  )
Description: An efficient implementation of the 'TreeSHAP' algorithm
    introduced by Lundberg et al., (2020) <doi:10.1038/s42256-019-0138-9>.
    It is capable of calculating SHAP (SHapley Additive exPlanations) values 
    for tree-based models in polynomial time.  Currently supported models include 
    'gbm', 'randomForest', 'ranger', 'xgboost', 'lightgbm'.
License: GPL-3
URL: https://modeloriented.github.io/treeshap/,
        https://github.com/ModelOriented/treeshap
BugReports: https://github.com/ModelOriented/treeshap/issues
Depends: R (>= 4.1.0)
Imports: data.table, ggplot2, Rcpp
Suggests: gbm, jsonlite, lightgbm, gpboost, randomForest, ranger,
        scales, survival, testthat, xgboost
LinkingTo: Rcpp
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-24 07:49:57 UTC; mayer
Author: Michael Mayer [aut, cre],
  Konrad Komisarczyk [aut],
  Pawel Kozminski [aut],
  Szymon Maksymiuk [aut] (ORCID: <https://orcid.org/0000-0002-3120-1601>),
  Lorenz A. Kapsner [ctb] (ORCID:
    <https://orcid.org/0000-0003-1866-860X>),
  Mikolaj Spytek [ctb] (ORCID: <https://orcid.org/0000-0001-7111-2286>),
  Mateusz Krzyzinski [ctb] (ORCID:
    <https://orcid.org/0000-0001-6143-488X>),
  Przemyslaw Biecek [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-8423-1823>)
Maintainer: Michael Mayer <mayermichael79@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-24 09:20:03 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-04-24 09:34:45 UTC; unix
Archs: treeshap.so.dSYM

Package: treeshap
Title: Compute SHAP Values for Your Tree-Based Models Using the
        'TreeSHAP' Algorithm
Version: 0.4.0
Authors@R: c(
    person("Michael", "Mayer", , "mayermichael79@gmail.com", role = c("aut", "cre")),
    person("Konrad", "Komisarczyk", email = "komisarczykkonrad@gmail.com", role = "aut"),
    person("Pawel", "Kozminski", email = "pkozminski99@gmail.com", role = "aut"),
    person("Szymon", "Maksymiuk", role = "aut", comment = c(ORCID = "0000-0002-3120-1601")),
    person("Lorenz A.", "Kapsner", role = "ctb", comment = c(ORCID = "0000-0003-1866-860X")),
    person("Mikolaj", "Spytek", role = "ctb", comment = c(ORCID = "0000-0001-7111-2286")),
    person("Mateusz", "Krzyzinski", email = "krzyzinskimateusz23@gmail.com", role = "ctb", comment = c(ORCID = "0000-0001-6143-488X")),
    person("Przemyslaw", "Biecek", email = "przemyslaw.biecek@gmail.com", role = c("aut", "cph"), comment = c(ORCID = "0000-0001-8423-1823"))
  )
Description: An efficient implementation of the 'TreeSHAP' algorithm
    introduced by Lundberg et al., (2020) <doi:10.1038/s42256-019-0138-9>.
    It is capable of calculating SHAP (SHapley Additive exPlanations) values 
    for tree-based models in polynomial time.  Currently supported models include 
    'gbm', 'randomForest', 'ranger', 'xgboost', 'lightgbm'.
License: GPL-3
URL: https://modeloriented.github.io/treeshap/,
        https://github.com/ModelOriented/treeshap
BugReports: https://github.com/ModelOriented/treeshap/issues
Depends: R (>= 4.1.0)
Imports: data.table, ggplot2, Rcpp
Suggests: gbm, jsonlite, lightgbm, gpboost, randomForest, ranger,
        scales, survival, testthat, xgboost
LinkingTo: Rcpp
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-24 07:49:57 UTC; mayer
Author: Michael Mayer [aut, cre],
  Konrad Komisarczyk [aut],
  Pawel Kozminski [aut],
  Szymon Maksymiuk [aut] (ORCID: <https://orcid.org/0000-0002-3120-1601>),
  Lorenz A. Kapsner [ctb] (ORCID:
    <https://orcid.org/0000-0003-1866-860X>),
  Mikolaj Spytek [ctb] (ORCID: <https://orcid.org/0000-0001-7111-2286>),
  Mateusz Krzyzinski [ctb] (ORCID:
    <https://orcid.org/0000-0001-6143-488X>),
  Przemyslaw Biecek [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-8423-1823>)
Maintainer: Michael Mayer <mayermichael79@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-24 09:20:03 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 12:35:43 UTC; unix
Archs: treeshap.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(predict,model_unified)
S3method(print,model_unified)
S3method(print,model_unified_multioutput)
S3method(print,treeshap)
S3method(print,treeshap_multioutput)
S3method(treeshap,model_unified)
S3method(treeshap,model_unified_multioutput)
S3method(unify,default)
S3method(unify,gbm)
S3method(unify,gpb.Booster)
S3method(unify,lgb.Booster)
S3method(unify,randomForest)
S3method(unify,ranger)
S3method(unify,xgb.Booster)
export(colors_breakdown_drwhy)
export(colors_discrete_drwhy)
export(gbm.unify)
export(gpboost.unify)
export(is.model_unified)
export(is.treeshap)
export(lightgbm.unify)
export(plot_contribution)
export(plot_feature_dependence)
export(plot_feature_importance)
export(plot_interaction)
export(randomForest.unify)
export(ranger.unify)
export(ranger_surv.unify)
export(set_reference_dataset)
export(theme_drwhy)
export(theme_drwhy_vertical)
export(treeshap)
export(unify)
export(xgboost.unify)
import(data.table)
import(ggplot2)
importFrom(Rcpp,sourceCpp)
importFrom(graphics,text)
importFrom(stats,reorder)
importFrom(stats,stepfun)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
useDynLib(treeshap)

#' SHAP value based Feature Dependence plot
#'
#' Depending on the value of a variable: how does it contribute into the prediction?
#'
#' @param treeshap A treeshap object produced with the \code{\link{treeshap}} function. \code{\link{treeshap.object}}.
#' @param variable name or index of variable for which feature dependence will be plotted.
#' @param title the plot's title, by default \code{'Feature Dependence'}.
#' @param subtitle the plot's subtitle. By default no subtitle.
#'
#' @return a \code{ggplot2} object
#'
#' @export
#'
#' @seealso
#' \code{\link{treeshap}} for calculation of SHAP values
#'
#' \code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_interaction}}
#'
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE) &&
#'  requireNamespace("scales", quietly = TRUE)) {
#'   library(xgboost)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'   xgb_model <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 3,
#'    nrounds = 20,
#'    nthreads = 1
#'   )
#'   unified_model <- xgboost.unify(xgb_model, as.matrix(data))
#'   x <- head(data, 100)
#'   shaps <- treeshap(unified_model, x)
#'   plot_feature_dependence(shaps, variable = "overall")
#' }}
plot_feature_dependence <- function(treeshap, variable,
                                    title = "Feature Dependence", subtitle = NULL) {
  if (!requireNamespace("scales", quietly = TRUE)) {
    stop(
      "Package \"scales\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }

  shaps <- treeshap$shaps
  x <- treeshap$observations

  # argument check
  if (!is.treeshap(treeshap)) {
    stop("treeshap parameter has to be correct object of class treeshap. Produce it using treeshap function.")
  }

  if (is.character(variable)) {
    if (!(variable %in% colnames(shaps))) {
      stop("Incorrect variable or shaps dataframe, variable should be one of variables in the shaps dataframe.")
    }
    if (!(variable %in% colnames(shaps))) {
      stop("Incorrect variable or x dataframe, varaible should be one of variables in the shaps dataframe.")
    }
  } else if (is.numeric(variable) && (length(variable) == 1)) {
    if (!all(colnames(shaps) == colnames(x))) {
      stop("shaps and x should have the same column names.")
    }
    if (!(variable %in% 1:ncol(shaps))) {
      stop("variable is an incorrect number.")
    }
    variable <- colnames(shaps)[variable]
  } else {
    stop("variable is of incorrect type.")
  }

  df <- data.frame(var_value = x[, variable], shap_value = shaps[, variable ])
  p <- ggplot(df, aes(x = var_value, y = shap_value)) +
    geom_point()

  p +
    theme_drwhy() +
    xlab(variable) + ylab(paste0("SHAP value for ", variable)) +
    labs(title = title, subtitle = subtitle) +
    scale_y_continuous(labels = scales::comma)
}

#' @import ggplot2
#' @import data.table
#' @importFrom stats reorder
#' @importFrom stats stepfun
#' @importFrom graphics text
#' @importFrom utils setTxtProgressBar txtProgressBar
#' @importFrom Rcpp sourceCpp
#' @useDynLib treeshap
NULL

#% https://community.rstudio.com/t/how-to-solve-no-visible-binding-for-global-
#% variable-note/28887
utils::globalVariables(
  c(
    "Feature", "split_index", "tree_index", ".",
    "node_parent", "default_left", "decision_type",
    "position", "cumulative", "prev", "text",
    "contribution", "var_value", "shap_value", "reorder",
    "variable", "importance", "Tree", "Node", "Missing",
    "Cover", "Yes", "No", 'Prediction', 'Decision.type'
  )
)

#' Calculate SHAP values of a tree ensemble model.
#'
#' Calculate SHAP values and optionally SHAP Interaction values.
#'
#'
#' @param unified_model Unified data.frame representation of the model created with a (model).unify function. A \code{\link{model_unified.object}} object.
#' @param x Observations to be explained. A \code{data.frame} or \code{matrix} object with the same columns as in the training set of the model. Keep in mind that objects different than \code{data.frame} or plain \code{matrix} will cause an error or unpredictable behavior.
#' @param interactions Whether to calculate SHAP interaction values. By default is \code{FALSE}. Basic SHAP values are always calculated.
#' @param verbose Whether to print progress bar to the console. Should be logical. Progress bar will not be displayed on Windows.
#'
#' @return A \code{\link{treeshap.object}} object (for single-output models) or \code{\link{treeshap_multioutput.object}}, which is a list of \code{\link{treeshap.object}} objects (for multi-output models). SHAP values can be accessed from \code{\link{treeshap.object}} with \code{$shaps}, and interaction values can be accessed with \code{$interactions}.
#'
#'
#' @export
#'
#' @seealso
#' \code{\link{xgboost.unify}} for \code{XGBoost models}
#' \code{\link{lightgbm.unify}} for \code{LightGBM models}
#' \code{\link{gpboost.unify}} for \code{GPBoost models}
#' \code{\link{gbm.unify}} for \code{GBM models}
#' \code{\link{randomForest.unify}} for \code{randomForest models}
#' \code{\link{ranger.unify}} for \code{ranger models}
#' \code{\link{ranger_surv.unify}} for \code{ranger survival models}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   library(xgboost)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'
#'   # calculating simple SHAP values
#'   xgb_model <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 3,
#'    nrounds = 20,
#'    nthreads = 1
#'   )
#'   unified_model <- xgboost.unify(xgb_model, as.matrix(data))
#'   treeshap1 <- treeshap(unified_model, head(data, 3))
#'   plot_contribution(treeshap1, obs = 1)
#'   treeshap1$shaps
#'
#'   # It's possible to calcualte explanation over different part of the data set
#'
#'   unified_model_rec <- set_reference_dataset(unified_model, data[1:1000, ])
#'   treeshap_rec <- treeshap(unified_model, head(data, 3))
#'   plot_contribution(treeshap_rec, obs = 1)
#'
#'   # calculating SHAP interaction values
#'   xgb_model2 <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 7,
#'    nrounds = 10,
#'    nthreads = 1
#'   )
#'   unified_model2 <- xgboost.unify(xgb_model2, as.matrix(data))
#'   treeshap2 <- treeshap(unified_model2, head(data, 3), interactions = TRUE)
#'   treeshap2$interactions
#' }}
#'
treeshap <- function(unified_model, x, interactions = FALSE, verbose = TRUE) {
  UseMethod("treeshap", unified_model)
}

#' @export
treeshap.model_unified <- function(unified_model, x, interactions = FALSE, verbose = TRUE){
  model <- unified_model$model
  # argument check
  if (!("matrix" %in% class(x) | "data.frame" %in% class(x))) {
    stop("x parameter has to be data.frame or matrix.")
  }

  if (!is.model_unified(unified_model)) {
    stop("unified_model parameter has to of class model_unified. Produce it using *.unify function.")
  }

  if (!attr(unified_model, "missing_support") & any(is.na(x))) {
    stop("Given model does not work with missing values. Dataset x should not contain missing values.")
  }

  x <- x[,colnames(x) %in% unified_model$feature_names]

  if (!all(model$Feature %in% c(NA, colnames(x)))) {
    stop("Dataset x does not contain all features occurring in the model.")
  }

  if (attr(unified_model, "model") == "LightGBM" & !is.data.frame(x)) {
    stop("For LightGBM models data.frame object is required as x parameter. Please convert.")
  }

  if ((!is.numeric(verbose) & !is.logical(verbose)) | is.null(verbose)) {
    warning("Incorrect verbose argument, setting verbose = FALSE (progress will not be printed).")
    verbose <- FALSE
  }
  verbose <- verbose[1] > 0 # so verbose = numeric will work too
  x <- as.data.frame(x)

  # adapting model representation to C++ and extracting from dataframe to vectors
  roots <- which(model$Node == 0) - 1
  yes <- model$Yes - 1
  no <- model$No - 1
  missing <- model$Missing - 1
  feature <- match(model$Feature, colnames(x)) - 1
  split <- model$Split
  decision_type <- unclass(model$Decision.type)
  is_leaf <- is.na(model$Feature)
  value <- model$Prediction
  cover <- model$Cover

  x2 <- as.data.frame(sapply(x, as.numeric))
  if (nrow(x) > 1) x2 <- t(x2) # transposed to be able to pick a observation with [] operator in Rcpp
  is_na <- is.na(x2) # needed, because dataframe passed to cpp somehow replaces missing values with random values

  # calculating SHAP values
  if (interactions) {
    result <- treeshap_interactions_cpp(x2, is_na,
                                        roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover,
                                        verbose)
    shaps <- result$shaps
    interactions_array <- array(result$interactions,
                                dim = c(ncol(x), ncol(x), nrow(x)),
                                dimnames = list(colnames(x), colnames(x), rownames(x)))
  } else {
    shaps <- treeshap_cpp(x2, is_na,
                          roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover,
                          verbose)
    interactions_array <- NULL
  }

  dimnames(shaps) <- list(rownames(x), colnames(x))
  treeshap_obj <- list(shaps = as.data.frame(shaps), interactions = interactions_array,
                       unified_model = unified_model, observations = x)
  class(treeshap_obj) <- "treeshap"
  return(treeshap_obj)
}


#' @export
treeshap.model_unified_multioutput <- function(unified_model, x, interactions = FALSE, verbose = TRUE){
  treeshaps_objects <- lapply(unified_model,
                              treeshap.model_unified,
                              x = x,
                              interactions = interactions,
                              verbose = verbose)
  class(treeshaps_objects) <- "treeshap_multioutput"
  return(treeshaps_objects)
}


#' treeshap results
#'
#' \code{treeshap} object produced by \code{treeshap} function.
#'
#' @return List consisting of four elements:
#' \describe{
#'   \item{shaps}{A \code{data.frame} with M columns, X rows (M - number of features, X - number of explained observations). Every row corresponds to SHAP values for a observation. }
#'   \item{interactions}{An \code{array} with dimensions (M, M, X) (M - number of features, X - number of explained observations). Every \code{[, , i]} slice is a symmetric matrix - SHAP Interaction values for a observation. \code{[a, b, i]} element is SHAP Interaction value of features \code{a} and \code{b} for observation \code{i}. Is \code{NULL} if interactions where not calculated (parameter \code{interactions} set \code{FALSE}.) }
#'   \item{unified_model}{An object of type \code{\link{model_unified.object}}. Unified representation of a model for which SHAP values were calculated. It is used by some of the plotting functions.}
#'   \item{observations}{Explained dataset. \code{data.frame} or \code{matrix}. It is used by some of the plotting functions.}
#' }
#'
#'
#' @seealso
#' \code{\link{treeshap}},
#'
#' \code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}, \code{\link{plot_interaction}}
#'
#'
#' @name treeshap.object
NULL


#' treeshap results for multi-output model
#'
#' \code{treeshap_multioutput} object produced by \code{treeshap} function.
#'
#' @return List consisting of \code{treeshap} objects, one for each individual output of a model. For survival models, the list is named using the time points, for which TreeSHAP values are calculated.
#'
#'
#' @seealso
#' \code{\link{treeshap}},
#'
#' \code{\link{treeshap.object}}
#'
#'
#' @name treeshap_multioutput.object
NULL


#' Prints treeshap objects
#'
#' @param x a treeshap object
#' @param ... other arguments
#'
#' @return No return value, called for printing
#'
#' @export
#'
print.treeshap <- function(x, ...){
  print(x$shaps)
  if (!is.null(x$interactions)) {
    print(x$interactions)
  }
  return(invisible(NULL))
}


#' Prints treeshap_multioutput objects
#'
#' @param x a treeshap_multioutput object
#' @param ... other arguments
#'
#' @return No return value, called for printing
#'
#' @export
#'
print.treeshap_multioutput <- function(x, ...){
  output_names <- names(x)
  lapply(output_names, function(output_name){
    cat(paste("-> for output:", output_name, "\n"))
    print(x[[output_name]])
    cat("\n")
  })
  return(invisible(NULL))
}


#' Check whether object is a valid treeshap object
#'
#' Does not check correctness of result, only basic checks
#'
#' @param x an object to check
#'
#' @return boolean
#'
#' @export
#'
is.treeshap <- function(x) {
  # class checks
  ("treeshap" %in% class(x)) &
    (is.data.frame(x$shaps)) &
    (is.null(x$interactions) | is.array(x$interactions)) &
    (is.model_unified(x$unified_model)) &
    (is.data.frame(x$observations) | is.matrix(x$observations)) &
    # dim checks
    (all(nrow(x$observations) == nrow(x$shaps)) & all(ncol(x$observations) == ncol(x$shaps))) &
    (is.null(x$interactions) | all(dim(x$interactions) == c(ncol(x$shaps), ncol(x$shaps), nrow(x$shaps)))) &
    # names check
    #all(rownames(x$observations) == rownames(x$shaps)) &
    all(colnames(x$observations) == colnames(x$shaps)) &
    (is.null(x$interactions) | all(dimnames(x$interactions)[[1]] == colnames(x$shaps)
                                & dimnames(x$interactions)[[2]] == colnames(x$shaps))) &
    #(is.null(x$interactions) | all(dimnames(x$interactions)[[3]] == rownames(x$shaps))) &
    # type check
    (is.null(x$interactions) | is.numeric(x$interactions)) &
    (is.numeric(as.matrix(x$shaps)))
}

if (requireNamespace("randomForest", quietly = TRUE)) {

  data_fifa <- fifa20$data[!colnames(fifa20$data) %in%
                            c('value_eur', 'gk_diving', 'gk_handling',
                              'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
  x <- na.omit(cbind(data_fifa, target = fifa20$target))

  rf_with_cat_model <- randomForest::randomForest(target~., data = x, maxnodes = 10, ntree = 10)

  x <- x[colnames(x) != 'work_rate']


  rf_num_model <- randomForest::randomForest(target~., data = x, maxnodes = 10, ntree = 10)


  test_that('randomForest.unify creates an object of the appropriate class', {
    expect_true(is.model_unified(randomForest.unify(rf_num_model, x)))
    expect_true(is.model_unified(unify(rf_num_model, x)))
  })

  test_that('randomForest.unify returns an object with correct attributes', {
    unified_model <- randomForest.unify(rf_num_model, x)

    expect_equal(attr(unified_model, "missing_support"), FALSE)
    expect_equal(attr(unified_model, "model"), "randomForest")
  })

  test_that('the randomForest.unify function returns data frame with columns of appropriate column', {
    unifier <- randomForest.unify(rf_num_model, x)$model
    expect_true(is.integer(unifier$Tree))
    expect_true(is.integer(unifier$Node))
    expect_true(is.character(unifier$Feature))
    expect_true(is.factor(unifier$Decision.type))
    expect_true(is.numeric(unifier$Split))
    expect_true(is.integer(unifier$Yes))
    expect_true(is.integer(unifier$No))
    expect_true(is.integer(unifier$Missing))
    expect_true(is.numeric(unifier$Prediction))
    expect_true(is.numeric(unifier$Cover))
  })

  test_that("shap calculates without an error", {
    unifier <- randomForest.unify(rf_num_model, x)
    expect_error(treeshap(unifier, x[1:3,], verbose = FALSE), NA)
  })

  test_that("randomForest: predictions from unified == original predictions", {
    unifier <- randomForest.unify(rf_num_model, x)
    obs <- x[1:16000, ]
    original <- stats::predict(rf_num_model, obs)
    names(original) <- NULL
    from_unified <- predict(unifier, obs)
    expect_true(all(abs((from_unified - original) / original) < 10**(-14)))
  })

  test_that("randomForest: mean prediction calculated using predict == using covers", {
    unifier <- randomForest.unify(rf_num_model, x)

    intercept_predict <- mean(predict(unifier, x))

    ntrees <- sum(unifier$model$Node == 0)
    leaves <- unifier$model[is.na(unifier$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that("randomForest: covers correctness", {
    unifier <- randomForest.unify(rf_num_model, x)

    roots <- unifier$model[unifier$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(x)))

    internals <- unifier$model[!is.na(unifier$model$Feature), ]
    yes_child_cover <- unifier$model[internals$Yes, ]$Cover
    no_child_cover <- unifier$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unifier$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[internals$Missing == internals$Yes | internals$Missing == internals$No] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })
}

## Small tests comparing TreeSHAP results to brutal implementation results
## Extensive testing of correctness is not possible due to complexity of brutal implementation



data <- fifa20$data[, 3:6] # limiting columns for faster exponential calculation
stopifnot(all(!is.na(data)))

data_na <- fifa20$data[, c(3:5, 9)] # limiting columns for faster exponential calculation
stopifnot(any(is.na(data_na)))

target <- fifa20$target

test_model <- function(max_depth, nrounds, model_name = "xgboost",
                      test_data = data, test_target = target) {
  if (model_name == "xgboost") {
    xgb_model <- xgboost::xgboost(
      x = test_data,
      y = test_target,
      objective = "reg:squarederror",
      max_depth = max_depth,
      nrounds = nrounds
    )
    return(xgboost.unify(xgb_model, test_data))
  } else if (model_name == "ranger") {
    if (any(is.na(test_data))) {
      stop("ranger does not work with NAa")
    }
    rf <- ranger::ranger(
      test_target ~ .,
      data = test_data,
      max.depth = max_depth,
      num.trees = nrounds
    )
    return(ranger.unify(rf, test_data))
  } else if (model_name == "randomForest") {
    if (any(is.na(test_data))) {
      stop("randomForest does not work with NAa")
    }
    rf <- randomForest::randomForest(
      test_target ~ .,
      data = test_data,
      maxnodes = 2**max_depth,
      ntree = nrounds
    )
    return(randomForest.unify(rf, test_data))
  } else if (model_name == "gbm") {
    gbm_data <- test_data
    gbm_data["gbm_target"] <- test_target
    gbm_model <- gbm::gbm(
      formula = gbm_target ~ .,
      data = gbm_data,
      distribution = "gaussian",
      n.trees = nrounds,
      interaction.depth = max_depth,
      n.cores = 1
    )
    return(gbm.unify(gbm_model, gbm_data))
  } else if (model_name == "lightgbm") {
    param_lgbm <- list(
      objective = "regression",
      max_depth = max_depth,
      force_row_wise = TRUE
    )
    x <- lightgbm::lgb.Dataset(
      as.matrix(test_data),
      label = as.matrix(test_target)
    )
    lgb_data <- lightgbm::lgb.Dataset.construct(x)
    lgb_model <- lightgbm::lightgbm(
      data = lgb_data,
      params = param_lgbm,
      nrounds = nrounds,
      verbose = -1,
      num_threads = 1
    )
    return(lightgbm.unify(lgb_model, as.matrix(test_data)))
  }
}


## Implementation of exponential complexity SHAP calculation

# functions wrapping tree structure
is_leaf <- function(model, j) (is.na(model$Feature[j]))
leaf_value <- function(model, j) {
  stopifnot(is_leaf(model, j))
  model$Prediction[j]
}
feature <- function(model, j) (model$Feature[j])
lesser <- function(model, j) (model$Yes[j])
greater <- function(model, j) (model$No[j])
missing <- function(model, j) (model$Missing[j])
threshold <- function(model, j) (model$Split[j])
cover <- function(model, j) (model$Cover[j])
extract_tree_root <- function(model, i) (which((model$Tree == i) & (model$Node == 0)))
tree_ids <- function(model) (unique(model$Tree))

# function estimating E[f(x) | x_S]
expvalue <- function(tree, root, x, S) {
  n <- ncol(x)
  G <- function(j, w) {
    if (is_leaf(tree, j)) {
      w * leaf_value(tree, j)
    } else {
      aj <- lesser(tree, j)
      bj <- greater(tree, j)
      cj <- missing(tree, j)
      stopifnot(length(aj) == 1)
      stopifnot(length(bj) == 1)
      stopifnot(length(cj) == 1)
      if (feature(tree, j) %in% S) {
        if (is.na(x[[feature(tree, j)]])) {
          if (!is.na(cj)) {
            G(cj, w)
          } else {
            stop("model does not work with NAs!")
          }
        } else if (x[[feature(tree, j)]] <= threshold(tree, j)) {
          G(aj, w)
        } else {
          G(bj, w)
        }
      } else {
        if (is.na(cj) | cj == aj | cj == bj) {
          G(aj, w * cover(tree, aj) / cover(tree, j)) + G(bj, w * cover(tree, bj) / cover(tree, j))
        } else {
          G(aj, w * cover(tree, aj) / cover(tree, j)) + G(bj, w * cover(tree, bj) / cover(tree, j)) + G(cj, w * cover(tree, cj) / cover(tree, j))
        }
      }
    }
  }
  G(root, 1)
}

# Function coppied form rje package ver 1.10.16
powerset <- function (x, m, rev = FALSE) {
  if (base::missing(m)) m = length(x)
  if (m == 0) return(list(x[c()]))

  out = list(x[c()])
  if (length(x) == 1)
    return(c(out, list(x)))
  for (i in seq_along(x)) {
    if (rev)
      out = c(lapply(out[lengths(out) < m], function(y) c(y, x[i])), out)
    else out = c(out, lapply(out[lengths(out) < m], function(y) c(y, x[i])))
  }
  out
}

# exponential calculation of SHAP Values
shap_exponential <- function(unified_model, x) {
  model <- unified_model$model

  shaps <- data.frame()
  for (row in 1:nrow(x)) {
    m <- ncol(x)
    shaps.row <- rep(0, times = m)
    for (t in tree_ids(model)) {
      root <- extract_tree_root(model, t)
      for (var_id in 1:m) {
        oth_vars <- colnames(x)[-var_id]
        var <- colnames(x)[var_id]
        sets <- powerset(oth_vars)
        for (S in sets) {
          f_without <- expvalue(model, root, x[row, ], S)
          f_with <- expvalue(model, root, x[row, ], c(S, var))
          size <- length(S)
          weight <- factorial(size) * factorial(m - size - 1) / factorial(m)
          shaps.row[var_id] <- shaps.row[var_id] + (f_with - f_without) * weight
        }
      }
    }
    shaps <- rbind(shaps, as.data.frame(matrix(shaps.row, nrow = 1)))
  }

  colnames(shaps) <- colnames(x)
  shaps
}

# exponential calculation of SHAP Interaction Values
shap_interactions_exponential <- function(unified_model, x) {
  model <- unified_model$model

  shaps <- as.matrix(shap_exponential(unified_model, x))
  interactions_array <- array(0,
                              dim = c(ncol(x), ncol(x), nrow(x)),
                              dimnames = list(colnames(x), colnames(x), c()))
  for (row in 1:nrow(x)) {
    m <- ncol(x)
    for (t in tree_ids(model)) {
      root <- extract_tree_root(model, t)
      for (var1_id in 1:m) {
        for (var2_id in 1:m) {
          if (var1_id < var2_id) {
            oth_vars <- colnames(x)[-c(var1_id, var2_id)]
            var1 <- colnames(x)[var1_id]
            var2 <- colnames(x)[var2_id]
            sets <- powerset(oth_vars)
            for (S in sets) {
              f_without_both <- expvalue(model, root, x[row, ], S)
              f_without_1 <- expvalue(model, root, x[row, ], c(S, var2))
              f_without_2 <- expvalue(model, root, x[row, ], c(S, var1))
              f_with <- expvalue(model, root, x[row, ], c(S, var1, var2))
              size <- length(S)
              weight <- factorial(size) * factorial(m - size - 2) / (2 * factorial(m - 1))
              sum <- weight * (f_with + f_without_both - f_without_1 - f_without_2)

              interactions_array[var1_id, var2_id, row] <- interactions_array[var1_id, var2_id, row] + sum
              interactions_array[var2_id, var1_id, row] <- interactions_array[var2_id, var1_id, row] + sum
            }
          }
        }
      }
    }

    # filling the diagonal
    row_sums <- apply(interactions_array[, , row], 2, sum)
    diag(interactions_array[, , row]) <- shaps[row, ] - row_sums
  }

  interactions_array
}



treeshap_correctness_test <- function(max_depth, nrounds, nobservations,
                                      model_name = "xgboost", test_data = data, test_target = target) {
  model <- test_model(max_depth, nrounds, model_name, test_data, test_target)
  set.seed(21)
  rows <- sample(nrow(test_data), nobservations)
  shaps_exp <- shap_exponential(model, test_data[rows, ])
  treeshap_res <- treeshap(model, test_data[rows, ], verbose = FALSE)
  shaps_treeshap <- treeshap_res$shaps

  precision <- 4
  is.treeshap(treeshap_res) & all(round(shaps_exp, precision) == round(shaps_treeshap, precision))
}

interactions_correctness_test <- function(max_depth, nrounds, nobservations,
                                          model_name = "xgboost", test_data = data, test_target = target) {
  model <- test_model(max_depth, nrounds, model_name, test_data, test_target)
  set.seed(21)
  rows <- sample(nrow(test_data), nobservations)
  interactions_exp <- shap_interactions_exponential(model, test_data[rows, ])
  treeshap_res <- treeshap(model, test_data[rows, ], interactions = TRUE, verbose = FALSE)
  interactions_treeshap <- treeshap_res$interactions

  precision_relative <- 1e-08
  precision_absolute <- 1e-08
  relative_error <- abs((interactions_exp - interactions_treeshap) / interactions_exp) < precision_relative
  absolute_error <- abs(interactions_exp - interactions_treeshap) < precision_absolute
  error <- relative_error | absolute_error
  is.treeshap(treeshap_res) & all(error)
}

shaps_sum_test <- function(max_depth, nrounds, nobservations,
                          model_type = "xgboost", test_data = data, test_target = target, precision = 1e-12) {
  model <- test_model(max_depth, nrounds, model_type, test_data, test_target)
  set.seed(21)
  rows <- sample(nrow(test_data), nobservations)

  ntrees <- sum(model$model$Node == 0)
  leaves <- model$model[is.na(model$model$Feature), ]
  intercept <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees
  prediction <- predict(model, test_data[rows, ])
  prediction_deviation <- prediction - intercept

  treeshap_res <- treeshap(model, test_data[rows, ], interactions = TRUE, verbose = FALSE)

  basic_shaps_sum <- apply(treeshap_res$shaps, 1, sum)
  expect_true(all(abs((prediction_deviation - basic_shaps_sum) / prediction_deviation) < precision))

  interactions_sum <- apply(treeshap_res$interactions, 3, sum)
  expect_true(all(abs((prediction_deviation - interactions_sum) / prediction_deviation) < precision))
}


if (requireNamespace("lightgbm", quietly = TRUE)) {

  test_that("treeshap function checks", {
    library(lightgbm)
    param_lgbm <- list(objective = "regression", max_depth = 2,  force_row_wise = TRUE)
    data_fifa <- fifa20$data[!colnames(fifa20$data) %in%
                c('work_rate', 'value_eur', 'gk_diving', 'gk_handling',
                'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]

    data_df <- as.matrix(na.omit(cbind(data_fifa, fifa20$target)))
    sparse_data <- data_df[,-ncol(data_df)]
    x <- lightgbm::lgb.Dataset(sparse_data, label = data_df[,ncol(data_df)])
    lgb_data <- lightgbm::lgb.Dataset.construct(x)
    lgb_model <- lightgbm::lightgbm(data = lgb_data,
                                    params = param_lgbm,
                                    verbose = -1,
                                    num_threads = 1)
    unified_model <- lightgbm.unify(lgb_model, sparse_data)
    expect_error(treeshap(unified_model, sparse_data[1:2,], verbose = FALSE))
  })
}


if (requireNamespace("xgboost", quietly = TRUE)) {
  test_that('treeshap correctness test 1 (xgboost, max_depth = 3, nrounds = 1, nobservations = 25)', {
    expect_true(treeshap_correctness_test(
      max_depth = 3,
      nrounds = 1,
      nobservations = 25,
      model_name = "xgboost"
    ))
  })

  test_that('treeshap correctness test 2 (xgboost, max_depth = 12, nrounds = 3, nobservations = 5)', {
    expect_true(treeshap_correctness_test(
      max_depth = 12,
      nrounds = 3,
      nobservations = 5,
      model_name = "xgboost"
    ))
  })

  test_that('treeshap correctness test 3 (xgboost, max_depth = 7, nrounds = 7, nobservations = 3, with NAs)', {
    expect_true(treeshap_correctness_test(
      max_depth = 7,
      nrounds = 7,
      nobservations = 3,
      model_name = "xgboost",
      test_data = data_na
    ))
  })
}


if (requireNamespace("ranger", quietly = TRUE)) {
  test_that('treeshap correctness test 4 (ranger, max_depth = 5, nrounds = 7, nobservations = 5)', {
    expect_true(treeshap_correctness_test(
      max_depth = 5,
      nrounds = 7,
      nobservations = 5,
      model_name = "ranger"
    ))
  })
}


if (requireNamespace("randomForest", quietly = TRUE)) {
  test_that('treeshap correctness test 5 (randomForest, max_depth = 3, nrounds = 7, nobservations = 5)', {
    expect_true(treeshap_correctness_test(
      max_depth = 3,
      nrounds = 7,
      nobservations = 5,
      model_name = "randomForest"
    ))
  })
}


if (requireNamespace("gbm", quietly = TRUE)) {
  test_that('treeshap correctness test 6 (gbm, max_depth = 3, nrounds = 7, nobservations = 5, with NAs)', {
    expect_true(treeshap_correctness_test(
      max_depth = 3,
      nrounds = 7,
      nobservations = 5,
      model_name = "gbm",
      test_data = data_na
    ))
  })
}


if (
  requireNamespace("lightgbm", quietly = TRUE) &&
    requireNamespace("jsonlite", quietly = TRUE)
) {
  test_that('treeshap correctness test 7 (lightgbm, max_depth = 3, nrounds = 7, nobservations = 5, with NAs)', {
    expect_true(treeshap_correctness_test(
      max_depth = 3,
      nrounds = 7,
      nobservations = 5,
      model_name = "lightgbm",
      test_data = data_na
    ))
  })
}


if (requireNamespace("xgboost", quietly = TRUE)) {
  test_that('interactions correctness test 1 (xgboost, max_depth = 3, nrounds = 1, nobservations = 25)', {
    expect_true(interactions_correctness_test(
      max_depth = 3,
      nrounds = 1,
      nobservations = 25,
      model_name = "xgboost"
    ))
  })

  test_that('interactions correctness test 2 (xgboost, max_depth = 12, nrounds = 3, nobservations = 5)', {
    expect_true(interactions_correctness_test(
      max_depth = 12,
      nrounds = 3,
      nobservations = 5,
      model_name = "xgboost"
    ))
  })

  test_that('interactions correctness test 3 (xgboost, max_depth = 7, nrounds = 4, nobservations = 2, with NAs)', {
    expect_true(interactions_correctness_test(
      max_depth = 7,
      nrounds = 4,
      nobservations = 2,
      model_name = "xgboost",
      test_data = data_na
    ))
  })
}

if (requireNamespace("ranger", quietly = TRUE)) {
  test_that('interactions correctness test 4 (ranger, max_depth = 6, nrounds = 4, nobservations = 2)', {
    expect_true(interactions_correctness_test(
      max_depth = 6,
      nrounds = 4,
      nobservations = 2,
      model_name = "ranger"
    ))
  })
}


if (requireNamespace("randomForest", quietly = TRUE)) {
  test_that('interactions correctness test 5 (randomForest, max_depth = 4, nrounds = 4, nobservations = 2)', {
    expect_true(interactions_correctness_test(
      max_depth = 4,
      nrounds = 4,
      nobservations = 2,
      model_name = "randomForest"
    ))
  })
}

if (requireNamespace("gbm", quietly = TRUE)) {
  test_that('interactions correctness test 6 (gbm, max_depth = 4, nrounds = 4, nobservations = 2, with NAs)', {
    expect_true(interactions_correctness_test(
      max_depth = 4,
      nrounds = 4,
      nobservations = 2,
      model_name = "gbm",
      test_data = data_na
    ))
  })
}


if (
  requireNamespace("lightgbm", quietly = TRUE) &&
    requireNamespace("jsonlite", quietly = TRUE)
) {
  test_that('interactions correctness test 7 (lightgbm, max_depth = 4, nrounds = 4, nobservations = 2, with NAs)', {
    expect_true(interactions_correctness_test(
      max_depth = 4,
      nrounds = 4,
      nobservations = 2,
      model_name = "lightgbm",
      test_data = data_na
    ))
  })
}


if (requireNamespace("xgboost", quietly = TRUE)) {
  test_that('xgboost: shaps sum up to prediction deviation (max_depth = 6, nrounds = 100, nobservations = 40, with NAs)', {
    shaps_sum_test(
      model_type = "xgboost",
      max_depth = 6,
      nrounds = 100,
      nobservations = 40,
      test_data = data_na
    )
  })
}


if (requireNamespace("ranger", quietly = TRUE)) {
  test_that('ranger: shaps sum up to prediction deviation (max_depth = 6, nrounds = 100, nobservations = 40)', {
    shaps_sum_test(
      model_type = "ranger",
      max_depth = 6,
      nrounds = 100,
      nobservations = 40
    )
  })

  test_that('randomForest: shaps sum up to prediction deviation (max_depth = 6, nrounds = 100, nobservations = 40)', {
    shaps_sum_test(
      model_type = "randomForest",
      max_depth = 6,
      nrounds = 100,
      nobservations = 40
    )
  })
}


if (requireNamespace("gbm", quietly = TRUE)) {
  test_that('gbm: shaps sum up to prediction deviation (max_depth = 6, nrounds = 40, nobservations = 40, with NAs)', {
    shaps_sum_test(
      model_type = "gbm",
      max_depth = 6,
      nrounds = 40,
      nobservations = 40,
      test_data = data_na
    )
  })
}

if (
  requireNamespace("lightgbm", quietly = TRUE) &&
  requireNamespace("jsonlite", quietly = TRUE)
) {
  test_that('lightgbm: shaps sum up to prediction deviation (max_depth = 6, nrounds = 100, nobservations = 40, with NAs)', {
    shaps_sum_test(
      model_type = "lightgbm",
      max_depth = 4,
      nrounds = 100,
      nobservations = 40,
      test_data = data_na
    )
  })
}

if (
  requireNamespace("survival", quietly = TRUE) &&
  requireNamespace("ranger", quietly = TRUE)
) {
  test_that('treeshap for multi-output model (survival ranger) works', {
    data_colon <- data.table::data.table(survival::colon)
    data_colon <- na.omit(data_colon[get("etype") == 2, ])
    surv_cols <- c("status", "time", "rx")

    feature_cols <- colnames(data_colon)[3:(ncol(data_colon) - 1)]

    x <- model.matrix(
      ~ -1 + .,
      data_colon[, .SD, .SDcols = setdiff(feature_cols, surv_cols[1:2])]
    )
    y <- survival::Surv(
      event = (data_colon[, get("status")] |>
        as.character() |>
        as.integer()),
      time = data_colon[, get("time")],
      type = "right"
    )

    ranger_num_model <- ranger::ranger(
      x = x,
      y = y,
      data = data_colon,
      max.depth = 10,
      num.trees = 10
    )

    unified_model_risk <- unify(ranger_num_model, x)
    times <- c(10, 50, 100)
    unified_model_survival <- unify(
      ranger_num_model,
      x,
      type = "survival",
      times = times
    )
    treeshaps_risk <- treeshap(unified_model_risk, x[1:2, ])
    treeshaps_survival <- treeshap(unified_model_survival, x[1:2, ])

    expect_s3_class(treeshaps_risk, "treeshap")
    expect_s3_class(treeshaps_survival, "treeshap_multioutput")
    expect_equal(length(treeshaps_survival), length(times))
    expect_equal(names(treeshaps_survival), as.character(times))
  })
}

if (requireNamespace("gbm", quietly = TRUE)) {

  x <- fifa20$data
  x['value_eur'] <- fifa20$target

  gbm_with_cat_model <- gbm::gbm(
    formula = value_eur ~ .,
    data = x,
    distribution = "laplace",
    n.trees = 10,
    interaction.depth = 2,
    n.cores = 1
  )

  x <- x[colnames(fifa20$data) != 'work_rate']

  gbm_num_model <- gbm::gbm(
    formula = value_eur ~ .,
    data = x,
    n.trees = 50,
    distribution = 'gaussian',
    n.cores = 1
  )

  test_that('gbm.unify returns an object of appropriate class', {
    expect_true(is.model_unified(gbm.unify(gbm_num_model, x)))
    expect_true(is.model_unified(unify(gbm_num_model, x)))
  })

  test_that('gbm.unify returns an object with correct attributes', {
    unified_model <- gbm.unify(gbm_num_model, x)

    expect_equal(attr(unified_model, "missing_support"), TRUE)
    expect_equal(attr(unified_model, "model"), "gbm")
  })

  test_that('the gbm.unify function does not support models with categorical features', {
    expect_error(
      gbm.unify(gbm_with_cat_model),
      "Models built on data with categorical features are not supported - please encode them before training."
    )
  })

  test_that('the gbm.unify function returns data frame with columns of appropriate column', {
    unifier <- gbm.unify(gbm_num_model, x)$model
    expect_true(is.integer(unifier$Tree))
    expect_true(is.integer(unifier$Node))
    expect_true(is.character(unifier$Feature))
    expect_true(is.factor(unifier$Decision.type))
    expect_true(is.numeric(unifier$Split))
    expect_true(is.integer(unifier$Yes))
    expect_true(is.integer(unifier$No))
    expect_true(is.integer(unifier$Missing))
    expect_true(is.numeric(unifier$Prediction))
    expect_true(is.numeric(unifier$Cover))
  })

  test_that("shap calculates without an error", {
    unifier <- gbm.unify(gbm_num_model, x)
    expect_error(treeshap(unifier, x[1:3, ], verbose = FALSE), NA)
  })

  test_that("gbm: mean prediction calculated using predict == using covers", {
    unifier <- gbm.unify(gbm_num_model, x)

    intercept_predict <- mean(predict(unifier, x))

    ntrees <- sum(unifier$model$Node == 0)
    leaves <- unifier$model[is.na(unifier$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) /
      sum(leaves$Cover) *
      ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that("gbm: covers correctness", {
    unifier <- gbm.unify(gbm_num_model, x)

    roots <- unifier$model[unifier$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(x)))

    internals <- unifier$model[!is.na(unifier$model$Feature), ]
    yes_child_cover <- unifier$model[internals$Yes, ]$Cover
    no_child_cover <- unifier$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unifier$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[
        internals$Missing == internals$Yes | internals$Missing == internals$No
      ] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })
}

if (requireNamespace("survival", quietly = TRUE) &&
  requireNamespace("ranger", quietly = TRUE)) {

  data_colon <- data.table::data.table(survival::colon)
  data_colon <- na.omit(data_colon[get("etype") == 2, ])
  surv_cols <- c("status", "time", "rx")

  feature_cols <- colnames(data_colon)[3:(ncol(data_colon) - 1)]

  x <- model.matrix(
    ~ -1 + .,
    data_colon[, .SD, .SDcols = setdiff(feature_cols, surv_cols[1:2])]
  )
  y <- survival::Surv(
    event = (data_colon[, get("status")] |>
              as.character() |>
              as.integer()),
    time = data_colon[, get("time")],
    type = "right"
  )

  set.seed(123)
  ranger_num_model <- ranger::ranger(
    x = x,
    y = y,
    data = data_colon,
    max.depth = 10,
    num.trees = 10
  )



  # to save some time for these tests, compute model here once:
  unified_model <- ranger_surv.unify(ranger_num_model, x)
  unified_model2 <- unify(ranger_num_model, x)


  test_that('ranger_surv.unify creates an object of the appropriate class', {
    expect_true(is.model_unified(unified_model))
    expect_true(is.model_unified(unified_model2))
  })

  test_that('ranger_surv.unify returns an object with correct attributes', {
    expect_equal(attr(unified_model, "missing_support"), FALSE)
    expect_equal(attr(unified_model, "model"), "ranger")
  })

  test_that('the ranger_surv.unify function returns data frame with columns of appropriate column', {
    unifier <- unified_model$model
    expect_true(is.integer(unifier$Tree))
    expect_true(is.integer(unifier$Node))
    expect_true(is.character(unifier$Feature))
    expect_true(is.factor(unifier$Decision.type))
    expect_true(is.numeric(unifier$Split))
    expect_true(is.integer(unifier$Yes))
    expect_true(is.integer(unifier$No))
    expect_true(all(is.na(unifier$Missing)))
    expect_true(is.numeric(unifier$Prediction))
    expect_true(is.numeric(unifier$Cover))
  })

  test_that("ranger_surv: shap calculates without an error", {
    expect_error(treeshap(unified_model, x[1:3,], verbose = FALSE), NA)
  })

  test_that("ranger_surv: predictions from unified == original predictions", {
    obs <- x[1:800, ]
    surv_preds <- stats::predict(ranger_num_model, obs)
    original <- rowSums(surv_preds$chf)
    from_unified <- predict(unified_model, obs)
    expect_true(all(abs((from_unified - original) / original) < 10**(-13)))
  })

  test_that("ranger_surv: mean prediction calculated using predict == using covers", {

    intercept_predict <- mean(predict(unified_model, x))

    ntrees <- sum(unified_model$model$Node == 0)
    leaves <- unified_model$model[is.na(unified_model$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that("ranger_surv: covers correctness", {

    roots <- unified_model$model[unified_model$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(x)))

    internals <- unified_model$model[!is.na(unified_model$model$Feature), ]
    yes_child_cover <- unified_model$model[internals$Yes, ]$Cover
    no_child_cover <- unified_model$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unified_model$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[internals$Missing == internals$Yes | internals$Missing == internals$No] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })


  # tests for ranger_surv.unify (type = "survival")
  # to save some time for these tests, compute model here once:
  unified_model <- ranger_surv.unify(ranger_num_model, x, type = "survival", times = c(10, 50, 100))
  unified_model2 <- unify(ranger_num_model, x, type = "survival", times = c(10, 50, 100))


  test_that('ranger_surv.unify (type = "survival") list names == unique.death.times', {
    expect_equal(names(unified_model), as.character(c(10, 50, 100)))
    expect_equal(names(unified_model2), as.character(c(10, 50, 100)))
  })

  test_that('ranger_surv.unify (type = "survival") creates an object of the appropriate class', {
    expect_s3_class(unified_model, "model_unified_multioutput")
    expect_s3_class(unified_model2, "model_unified_multioutput")
    lapply(unified_model, function(m) expect_true(is.model_unified(m)))
    lapply(unified_model2, function(m) expect_true(is.model_unified(m)))
  })

  test_that('ranger_surv.unify (type = "survival") returns an object with correct attributes', {
    m <- unified_model[[1]]
    expect_equal(attr(m, "missing_support"), FALSE)
    expect_equal(attr(m, "model"), "ranger")
  })

  test_that('the ranger_surv.unify (type = "survival") function returns data frame with columns of appropriate column', {
      unifier <- unified_model[[1]]$model
      expect_true(is.integer(unifier$Tree))
      expect_true(is.integer(unifier$Node))
      expect_true(is.character(unifier$Feature))
      expect_true(is.factor(unifier$Decision.type))
      expect_true(is.numeric(unifier$Split))
      expect_true(is.integer(unifier$Yes))
      expect_true(is.integer(unifier$No))
      expect_true(all(is.na(unifier$Missing)))
      expect_true(is.numeric(unifier$Prediction))
      expect_true(is.numeric(unifier$Cover))
  })

  test_that('ranger_surv.unify (type = "survival"): shap calculates without an error', {
    expect_error(treeshap(unified_model[[1]], x[1:3,], verbose = FALSE), NA)
  })

  test_that('ranger_surv.unify (type = "survival"): predictions from unified == original predictions', {
    for (t in names(unified_model)) {
      m <- unified_model[[t]]
      death_time <- as.integer(t)
      obs <- x[1:800, ]
      surv_preds <- stats::predict(ranger_num_model, obs)
      original <- surv_preds$survival[, which(surv_preds$unique.death.times == death_time)]
      from_unified <- predict(m, obs)
      # this is yet kind of strange that values differ so much
      expect_true(all(abs((from_unified - original) / original) < 8e-1))
    }
  })

  test_that('ranger_surv.unify (type = "survival"): mean prediction calculated using predict == using covers', {
    m <- unified_model[[1]]
    intercept_predict <- mean(predict(m, x))

    ntrees <- sum(m$model$Node == 0)
    leaves <- m$model[is.na(m$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that('ranger_surv.unify (type = "survival"): covers correctness', {
    for (m in unified_model) {
      roots <- m$model[m$model$Node == 0, ]
      expect_true(all(roots$Cover == nrow(x)))

      internals <- m$model[!is.na(m$model$Feature), ]
      yes_child_cover <- m$model[internals$Yes, ]$Cover
      no_child_cover <- m$model[internals$No, ]$Cover
      if (all(is.na(internals$Missing))) {
        children_cover <- yes_child_cover + no_child_cover
      } else {
        missing_child_cover <- m$model[internals$Missing, ]$Cover
        missing_child_cover[is.na(missing_child_cover)] <- 0
        missing_child_cover[internals$Missing == internals$Yes | internals$Missing == internals$No] <- 0
        children_cover <- yes_child_cover + no_child_cover + missing_child_cover
      }
      expect_true(all(internals$Cover == children_cover))
    }
  })
}

## plotting functions for treeshap package

#' SHAP value based Feature Importance plot
#'
#' This function plots feature importance calculated as means of absolute values of SHAP values of variables (average impact on model output magnitude).
#'
#' @param treeshap A treeshap object produced with the \code{\link{treeshap}} function. \code{\link{treeshap.object}}.
#' @param desc_sorting logical. Should the bars be sorted descending? By default TRUE.
#' @param max_vars maximum number of variables that shall be presented. By default all are presented.
#' @param title the plot's title, by default \code{'Feature Importance'}.
#' @param subtitle the plot's subtitle. By default no subtitle.
#'
#' @return a \code{ggplot2} object
#'
#' @export
#'
#' @seealso
#' \code{\link{treeshap}} for calculation of SHAP values
#'
#' \code{\link{plot_contribution}}, \code{\link{plot_feature_dependence}}, \code{\link{plot_interaction}}
#'
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE) &&
#'  requireNamespace("scales", quietly = TRUE)) {
#'   library(xgboost)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'   xgb_model <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 3,
#'    nrounds = 20,
#'    nthreads = 1
#'   )
#'   unified_model <- xgboost.unify(xgb_model, as.matrix(data))
#'   shaps <- treeshap(unified_model, as.matrix(head(data, 3)))
#'   plot_feature_importance(shaps, max_vars = 4)
#' }}
plot_feature_importance <- function(treeshap,
                                    desc_sorting = TRUE,
                                    max_vars = ncol(shaps),
                                    title = "Feature Importance",
                                    subtitle = NULL) {
  if (!requireNamespace("scales", quietly = TRUE)) {
    stop(
      "Package \"scales\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }
  shaps <- treeshap$shaps

  # argument check
  if (!is.treeshap(treeshap)) {
    stop("treeshap parameter has to be correct object of class treeshap. Produce it using treeshap function.")
  }

  if (!is.logical(desc_sorting)) {
    stop("desc_sorting is not logical.")
  }

  if (!is.numeric(max_vars)) {
    stop("max_vars is not numeric.")
  }

  if (max_vars > ncol(shaps)) {
    warning("max_vars exceeded number of explained variables. All variables will be shown.")
    max_vars <- ncol(shaps)
  }

  mean <- colMeans(abs(shaps))
  df <- data.frame(variable = factor(names(mean)), importance = as.vector(mean))
  df$variable <- reorder(df$variable, df$importance * ifelse(desc_sorting, 1, -1))
  df <- df[order(df$importance, decreasing = TRUE)[1:max_vars], ]

  p <- ggplot(df, aes(x = variable, y = importance)) +
    geom_bar(stat = "identity", fill = colors_discrete_drwhy(1))

  p + coord_flip() +
    theme_drwhy_vertical() +
    ylab("mean(|SHAP value|)") + xlab("") +
    labs(title = title, subtitle = subtitle) +
    scale_y_continuous(labels = scales::comma) +
    theme(legend.position = "none")
}

#' Predict
#'
#' Predict using unified_model representation.
#'
#' @param object Unified model representation of the model created with a (model).unify function. \code{\link{model_unified.object}}
#' @param x Observations to predict. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model.
#' @param ... other parameters
#'
#' @return a vector of predictions.
#'
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("gbm", quietly = TRUE)) {
#'   library(gbm)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   data['value_eur'] <- fifa20$target
#'   gbm_model <- gbm::gbm(
#'     formula = value_eur ~ .,
#'     data = data,
#'     distribution = "laplace",
#'     n.trees = 20,
#'     interaction.depth = 4,
#'     n.cores = 1
#'   )
#'   unified <- gbm.unify(gbm_model, data)
#'   predict(unified, data[2001:2005, ])
#' }}
predict.model_unified <- function(object, x, ...) {
  unified_model <- object
  model <- unified_model$model
  x <- as.data.frame(x)

  # argument check
  if (!is.model_unified(unified_model)) {
    stop("unified_model parameter has to of class model_unified. Produce it using *.unify function.")
  }

  if (!("matrix" %in% class(x) | "data.frame" %in% class(x))) {
    stop("x parameter has to be data.frame or matrix.")
  }

  if (!attr(unified_model, "missing_support") && any(is.na(x))) {
    stop("Given model does not work with missing values. Dataset x should not contain missing values.")
  }

  x <- x[,colnames(x) %in% unified_model$feature_names]

  if (!all(model$Feature %in% c(NA, colnames(x)))) {
    stop("Dataset does not contain all features occurring in the model.")
  }

  # adapting model representation to C++ and extracting from dataframe to vectors
  roots <- which(model$Node == 0) - 1
  yes <- model$Yes - 1
  no <- model$No - 1
  missing <- model$Missing - 1
  is_leaf <- is.na(model$Feature)
  feature <- match(model$Feature, colnames(x)) - 1
  split <- model$Split
  decision_type <- unclass(model$Decision.type)
  #stopifnot(levels(decision_type) == c("<=", "<"))
  #stopifnot(all(decision_type %in% c(1, 2, NA)))
  value <- model$Prediction

  n <- nrow(x)
  x <- as.data.frame(sapply(x, as.numeric))
  if (n > 1) x <- t(x)

  is_na <- is.na(x) # needed, because dataframe passed to cpp somehow replaces missing values with random values

  predict_cpp(x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type, value)
}

#' Unify randomForest model
#'
#' Convert your randomForest model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' Binary classification models with a target variable that is a factor with two levels, 0 and 1, are supported
#'
#' @param rf_model An object of \code{randomForest} class. At the moment, models built on data with categorical features
#' are not supported - please encode them before training.
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#'
#' @return a unified model representation - a \code{\link{model_unified.object}} object
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' @examples
#' if (requireNamespace("randomForest", quietly = TRUE)) {
#'   library(randomForest)
#'   data_fifa <- fifa20$data[
#'     !colnames(fifa20$data) %in%
#'       c(
#'         'work_rate',
#'         'value_eur',
#'         'gk_diving',
#'         'gk_handling',
#'         'gk_kicking',
#'         'gk_reflexes',
#'         'gk_speed',
#'         'gk_positioning'
#'       )
#'   ]
#'   data <- na.omit(cbind(data_fifa, target = fifa20$target))
#'
#'   rf <- randomForest::randomForest(
#'     target ~ .,
#'     data = data,
#'     maxnodes = 10,
#'     ntree = 10
#'   )
#'   unified_model <- randomForest.unify(rf, data)
#'   shaps <- treeshap(unified_model, data[1:2, ])
#'    # plot_contribution(shaps, obs = 1)
#' }
#'
#' @export
#'
randomForest.unify <- function(rf_model, data) {
  if (!inherits(rf_model, 'randomForest')) {
    stop('Object rf_model was not of class "randomForest"')
  }
  if (any(attr(rf_model$terms, "dataClasses")[-1] != "numeric")) {
    stop(
      'Models built on data with categorical features are not supported - please encode them before training.'
    )
  }
  n <- rf_model$ntree
  ret <- data.table()
  prediction <- NULL
  x <- lapply(1:n, function(tree) {
    tree_data <- as.data.table(randomForest::getTree(
      rf_model,
      k = tree,
      labelVar = TRUE
    ))
    tree_data <- tree_data[, prediction := as.numeric(prediction)]
    tree_data[, c(
      "left daughter",
      "right daughter",
      "split var",
      "split point",
      "prediction"
    )]
  })
  times_vec <- sapply(x, nrow)
  y <- rbindlist(x)
  y[, Tree := rep(0:(n - 1), times = times_vec)]
  y[, Node := unlist(lapply(times_vec, function(x) 0:(x - 1)))]
  setnames(y, c("Yes", "No", "Feature", "Split", "Prediction", "Tree", "Node"))
  y[, Feature := as.character(Feature)]
  y[, Yes := Yes - 1]
  y[, No := No - 1]
  y[y$Yes < 0, "Yes"] <- NA
  y[y$No < 0, "No"] <- NA
  y[, Missing := NA]
  y[, Missing := as.integer(Missing)] # seems not, but needed

  ID <- paste0(y$Node, "-", y$Tree)
  y$Yes <- match(paste0(y$Yes, "-", y$Tree), ID)
  y$No <- match(paste0(y$No, "-", y$Tree), ID)

  y$Cover <- 0

  y$Decision.type <- factor(
    x = rep("<=", times = nrow(y)),
    levels = c("<=", "<")
  )
  y[is.na(Feature), Decision.type := NA]

  # Here we lose "Quality" information
  y[!is.na(Feature), Prediction := NA]

  # treeSHAP assumes, that [prediction = sum of predictions of the trees]
  # in random forest [prediction = mean of predictions of the trees]
  # so here we correct it by adjusting leaf prediction values
  y[is.na(Feature), Prediction := Prediction / n]

  setcolorder(
    y,
    c(
      "Tree",
      "Node",
      "Feature",
      "Decision.type",
      "Split",
      "Yes",
      "No",
      "Missing",
      "Prediction",
      "Cover"
    )
  )
  feature_names <- rownames(rf_model$importance)
  data <- data[, colnames(data) %in% feature_names]

  ret <- list(
    model = as.data.frame(y),
    data = as.data.frame(data),
    feature_names = feature_names
  )
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- FALSE
  attr(ret, "model") <- "randomForest"
  return(set_reference_dataset(ret, as.data.frame(data)))
}

#' Unified model representation
#'
#' \code{model_unified} object produced by \code{*.unify} or \code{unify} function.
#'
#' @return List consisting of two elements:
#'
#'
#' \strong{model} - A \code{data.frame} representing model with following columns:
#'
#' \item{Tree}{0-indexed ID of a tree}
#' \item{Node}{0-indexed ID of a node in a tree. In a tree the root always has ID 0}
#' \item{Feature}{In case of an internal node - name of a feature to split on. Otherwise - NA}
#' \item{Decision.type}{A factor with two levels: "<" and "<=". In case of an internal node - predicate used for splitting observations. Otherwise - NA}
#' \item{Split}{For internal nodes threshold used for splitting observations. All observations that satisfy the predicate Decision.type(Split) ('< Split' / '<= Split') are proceeded to the node marked as 'Yes'. Otherwise to the 'No' node. For leaves - NA}
#' \item{Yes}{Index of a row containing a child Node. Thanks to explicit indicating the row it is much faster to move between nodes}
#' \item{No}{Index of a row containing a child Node}
#' \item{Missing}{Index of a row containing a child Node where are proceeded all observations with no value of the dividing feature}
#' \item{Prediction}{For leaves: Value of prediction in the leaf. For internal nodes: NA}
#' \item{Cover}{Number of observations seen by the internal node or collected by the leaf for the reference dataset}
#'
#' \strong{data} - Dataset used as a reference for calculating SHAP values. A dataset passed to the \code{*.unify}, \code{unify} or \code{\link{set_reference_dataset}} function with \code{data} argument. A \code{data.frame}.
#'
#'
#' Object has two also attributes set:
#' \item{\code{model}}{A string. By what package the model was produced.}
#' \item{\code{missing_support}}{A boolean. Whether the model allows missing values to be present in explained dataset.}
#'
#'
#' @seealso
#' \code{\link{unify}}
#'
#'
#' @name model_unified.object
#'
NULL


#' Unified model representations for multi-output model
#'
#' \code{model_unified_multioutput} object produced by \code{*.unify} or \code{unify} function.
#'
#' @return List consisting of \code{model_unified} objects, one for each individual output of a model. For survival models, the list is named using the time points, for which predictions are calculated.
#'
#' @seealso
#' \code{\link{unify}}
#'
#'
#' @name model_unified_multioutput.object
#'
NULL


#' Prints model_unified objects
#'
#' @param x a model_unified object
#' @param ... other arguments
#'
#' @return No return value, called for printing
#'
#' @export
#'
print.model_unified <- function(x, ...){
  print(x$model)
  return(invisible(NULL))
}


#' Prints model_unified_multioutput objects
#'
#' @param x a model_unified_multioutput object
#' @param ... other arguments
#'
#' @return No return value, called for printing
#'
#' @export
#'
print.model_unified_multioutput <- function(x, ...){
  output_names <- names(x)
  lapply(output_names, function(output_name){
    cat(paste("-> for output:", output_name, "\n"))
    print(x[[output_name]])
    cat("\n")
    })
  return(invisible(NULL))
}


#' Check whether object is a valid model_unified object
#'
#' Does not check correctness of representation, only basic checks
#'
#' @param x an object to check
#'
#' @return boolean
#'
#' @export
#'
is.model_unified <- function(x) {
  # class checks
  ("model_unified" %in% class(x)) &
    is.data.frame(x$data) &
    is.data.frame(x$model) &
    # attributes check
    is.character(attr(x, "model")) &
    is.logical(attr(x, "missing_support")) &
    # colnames check
    all(c("Tree", "Node", "Feature", "Decision.type", "Split", "Yes", "No", "Missing", "Prediction", "Cover") %in% colnames(x$model)) &
    # column types check
    is.numeric(x$model$Tree) &
    is.numeric(x$model$Node) &
    is.character(x$model$Feature) &
    is.factor(x$model$Decision.type) &
    all(levels(x$model$Decision.type) == c("<=", "<")) &
    all(unclass(x$model$Decision.type) %in% c(1, 2, NA)) &
    is.numeric(x$model$Split) &
    is.numeric(x$model$Yes) &
    is.numeric(x$model$No) &
    (!attr(x, "missing_support") | is.numeric(x$model$Missing)) &
    is.numeric(x$model$Prediction) &
    is.numeric(x$model$Cover)
}

if (requireNamespace("gpboost", quietly = TRUE)) {
  param_gpboost <- list(objective = "regression",
                        max_depth = 3,
                        force_row_wise = TRUE,
                        learning.rate = 0.1)

  data_fifa <- fifa20$data[!colnames(fifa20$data)%in%c('work_rate', 'value_eur', 'gk_diving', 'gk_handling', 'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
  data <- as.matrix(na.omit(data.table::as.data.table(cbind(data_fifa, fifa20$target))))
  sparse_data <- data[,-ncol(data)]
  x <- gpboost::gpb.Dataset(sparse_data, label = data[,ncol(data)])
  gpb_data <- gpboost::gpb.Dataset.construct(x)
  gpb_fifa <- gpboost::gpboost(data = gpb_data,
                        params = param_gpboost,
                        verbose = -1,
                        num_threads = 1)
  gpbtree <- gpboost::gpb.model.dt.tree(gpb_fifa)

  test_that('gpboost.unify returns an object with correct attributes', {
    unified_model <- gpboost.unify(gpb_fifa, sparse_data)

    expect_equal(attr(unified_model, "missing_support"), TRUE)
    expect_equal(attr(unified_model, "model"), "gpboost")
  })

  test_that('Columns after gpboost.unify are of appropriate type', {
    unified_model <- gpboost.unify(gpb_fifa, sparse_data)$model
    expect_true(is.integer(unified_model$Tree))
    expect_true(is.character(unified_model$Feature))
    expect_true(is.factor(unified_model$Decision.type))
    expect_true(is.numeric(unified_model$Split))
    expect_true(is.integer(unified_model$Yes))
    expect_true(is.integer(unified_model$No))
    expect_true(is.integer(unified_model$Missing))
    expect_true(is.numeric(unified_model$Prediction))
    expect_true(is.numeric(unified_model$Cover))
  })

  test_that('gpboost.unify creates an object of the appropriate class', {
    expect_true(is.model_unified(gpboost.unify(gpb_fifa, sparse_data)))
    expect_true(is.model_unified(unify(gpb_fifa, sparse_data)))
  })

  test_that('basic columns after gpboost.unify are correct', {
    unified_model <- gpboost.unify(gpb_fifa, sparse_data)$model
    expect_equal(gpbtree$tree_index, unified_model$Tree)
    to_test_features <- gpbtree[order(gpbtree$split_index), .(split_feature,split_index, threshold, leaf_count, internal_count),tree_index]
    expect_equal(to_test_features[!is.na(to_test_features$split_index),][['split_index']], unified_model[!is.na(unified_model$Feature),][['Node']])
    expect_equal(to_test_features[['split_feature']], unified_model[['Feature']])
    expect_equal(to_test_features[['threshold']], unified_model[['Split']])
    expect_equal(to_test_features[!is.na(internal_count),][['internal_count']], unified_model[!is.na(unified_model$Feature),][['Cover']])
  })

  test_that('connections between nodes and leaves after gpboost.unify are correct', {
    test_object <- as.data.table(gpboost.unify(gpb_fifa, sparse_data)$model)
    #Check if the sums of children's covers are correct
    expect_equal(test_object[test_object[!is.na(test_object$Yes)][['Yes']]][['Cover']] +
      test_object[test_object[!is.na(test_object$No)][['No']]][['Cover']], test_object[!is.na(Feature)][['Cover']])
    #check if default_left information is correctly used
    df_default_left <- gpbtree[default_left == "TRUE", c('tree_index', 'split_index')]
    test_object_actual_default_left <- test_object[Yes == Missing, c('Tree', 'Node')]
    colnames(test_object_actual_default_left) <- c('tree_index', 'split_index')
    attr(test_object_actual_default_left, 'model') <- NULL
    expect_equal(test_object_actual_default_left[order(tree_index, split_index)], df_default_left[order(tree_index, split_index)])
    #and default_left = FALSE analogically:
    df_default_right <- gpbtree[default_left != 'TRUE', c('tree_index', 'split_index')]
    test_object_actual_default_right <- test_object[No == Missing, c('Tree', 'Node')]
    colnames(test_object_actual_default_right) <- c('tree_index', 'split_index')
    attr(test_object_actual_default_right, 'model') <- NULL
    expect_equal(test_object_actual_default_right[order(tree_index, split_index)], df_default_right[order(tree_index, split_index)])
    #One more test with checking the usage of 'decision_type' column needed
  })

  # Function that return the predictions for sample observations indicated by vector contatining values -1, 0, 1, where -1 means
  # going to the 'Yes' Node, 1 - to the 'No' node and 0 - to the missing node. The vectors are randomly produced during executing
  # the function and should be passed to prepare_original_preds_ to save the conscistence. Later we can compare the 'predicted' values
  prepare_test_preds <- function(unify_out){
    stopifnot(all(c("Tree", "Node", "Feature", "Split", "Yes", "No", "Missing", "Prediction", "Cover") %in% colnames(unify_out)))
    test_tree <- unify_out[unify_out$Tree %in% 0:9,]
    test_tree[['node_row_id']] <- seq_len(nrow(test_tree))
    test_obs <- lapply(table(test_tree$Tree), function(y) sample(c(-1, 0, 1), y, replace = T))
    test_tree <- split(test_tree, test_tree$Tree)
    determine_val <- function(obs, tree){
      root_id <- tree[['node_row_id']][1]
      tree[,c('Yes', 'No', 'Missing')] <- tree[,c('Yes', 'No', 'Missing')] - root_id + 1
      i <- 1
      indx <- 1
      while(!is.na(tree$Feature[indx])) {
        indx <- ifelse(obs[i] == 0, tree$Missing[indx], ifelse(obs[i] < 0, tree$Yes[indx], tree$No[indx]))
        #if(length(is.na(tree$Feature[indx]))>1) {print(paste(indx, i)); print(tree); print(obs)}
        i <- i + 1
      }
      return(tree[['Prediction']][indx])
    }
    x = numeric()
    for(i in seq_along(test_obs)) {
      x[i] <- determine_val(test_obs[[i]], test_tree[[i]])

    }
    return(list(preds = x, test_obs = test_obs))
  }

  prepare_original_preds_gpb <- function(orig_tree, test_obs){
    test_tree <- orig_tree[orig_tree$tree_index %in% 0:9,]
    test_tree <- split(test_tree, test_tree$tree_index)
    stopifnot(length(test_tree) == length(test_obs))
    determine_val <- function(obs, tree){
      i <- 1
      indx <- 1
      while(!is.na(tree$split_feature[indx])) {
        children <- ifelse(is.na(tree$node_parent), tree$leaf_parent, tree$node_parent)
        if((obs[i] < 0) | (tree$default_left[indx] == 'TRUE' & obs[i] == 0)){
          indx <- which(tree$split_index[indx] == children)[1]
        }
        else if((obs[i] > 0) | (tree$default_left[indx] == 'FALSE' & obs[i] == 0)) {
          indx <- which(tree$split_index[indx] == children)[2]
        }
        else{
          stop('Error in the connections')
          indx <- 0
        }
        i <- i + 1
      }
      return(tree[['leaf_value']][indx])
    }
    y = numeric()
    for(i in seq_along(test_obs)) {
      y[i] <- determine_val(test_obs[[i]], test_tree[[i]])
    }
    return(y)
  }

  test_that('the connections between the nodes are correct', {
    # The test is passed only if the predictions for sample observations are equal in the first 10 trees of the ensemble
    x <- prepare_test_preds(gpboost.unify(gpb_fifa, sparse_data)$model)
    preds <- x[['preds']]
    test_obs <- x[['test_obs']]
    original_preds <- prepare_original_preds_gpb(gpbtree, test_obs)
    expect_equal(preds, original_preds)
  })

  test_that("gpboost: predictions from unified == original predictions", {
    unifier <- gpboost.unify(gpb_fifa, sparse_data)
    obs <- c(1:16000)
    original <- stats::predict(gpb_fifa, sparse_data[obs, ])
    from_unified <- predict(unifier, sparse_data[obs, ])
    expect_equal(from_unified, original)
    #expect_true(all(abs((from_unified - original) / original) < 10**(-14))) #not needed
  })

  test_that("gpboost: mean prediction calculated using predict == using covers", {
    unifier <- gpboost.unify(gpb_fifa, sparse_data)

    intercept_predict <- mean(predict(unifier, sparse_data))

    ntrees <- sum(unifier$model$Node == 0)
    leaves <- unifier$model[is.na(unifier$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that("gpboost: covers correctness", {
    unifier <- gpboost.unify(gpb_fifa, sparse_data)

    roots <- unifier$model[unifier$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(sparse_data)))

    internals <- unifier$model[!is.na(unifier$model$Feature), ]
    yes_child_cover <- unifier$model[internals$Yes, ]$Cover
    no_child_cover <- unifier$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unifier$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[internals$Missing == internals$Yes | internals$Missing == internals$No] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })
}

if (
  requireNamespace("lightgbm", quietly = TRUE) &&
  requireNamespace("jsonlite", quietly = TRUE)
) {
  param_lightgbm <- list(
    objective = "regression",
    max_depth = 3,
    force_row_wise = TRUE,
    learning.rate = 0.1
  )

  data_fifa <- fifa20$data[
    !colnames(fifa20$data) %in%
      c(
        'work_rate',
        'value_eur',
        'gk_diving',
        'gk_handling',
        'gk_kicking',
        'gk_reflexes',
        'gk_speed',
        'gk_positioning'
      )
  ]
  data <- as.matrix(na.omit(data.table::as.data.table(cbind(
    data_fifa,
    fifa20$target
  ))))
  sparse_data <- data[, -ncol(data)]
  x <- lightgbm::lgb.Dataset(sparse_data, label = data[, ncol(data)])
  lgb_data <- lightgbm::lgb.Dataset.construct(x)
  lgbm_fifa <- lightgbm::lightgbm(
    data = lgb_data,
    params = param_lightgbm,
    verbose = -1,
    num_threads = 1
  )
  lgbmtree <- lightgbm::lgb.model.dt.tree(lgbm_fifa)

  test_that('lightgbm.unify returns an object with correct attributes', {
    unified_model <- lightgbm.unify(lgbm_fifa, sparse_data)

    expect_equal(attr(unified_model, "missing_support"), TRUE)
    expect_equal(attr(unified_model, "model"), "LightGBM")
  })

  test_that('Columns after lightgbm.unify are of appropriate type', {
    unified_model <- lightgbm.unify(lgbm_fifa, sparse_data)$model
    expect_true(is.integer(unified_model$Tree))
    expect_true(is.integer(unified_model$Node))
    expect_true(is.character(unified_model$Feature))
    expect_true(is.factor(unified_model$Decision.type))
    expect_true(is.numeric(unified_model$Split))
    expect_true(is.integer(unified_model$Yes))
    expect_true(is.integer(unified_model$No))
    expect_true(is.integer(unified_model$Missing))
    expect_true(is.numeric(unified_model$Prediction))
    expect_true(is.numeric(unified_model$Cover))
  })

  test_that('lightgbm.unify creates an object of the appropriate class', {
    expect_true(is.model_unified(lightgbm.unify(lgbm_fifa, sparse_data)))
    expect_true(is.model_unified(unify(lgbm_fifa, sparse_data)))
  })

  test_that('basic columns after lightgbm.unify are correct', {
    unified_model <- lightgbm.unify(lgbm_fifa, sparse_data)$model
    expect_equal(lgbmtree$tree_index, unified_model$Tree)
    to_test_features <- lgbmtree[
      order(lgbmtree$split_index),
      .(split_feature, split_index, threshold, leaf_count, internal_count),
      tree_index
    ]
    expect_equal(
      to_test_features[!is.na(to_test_features$split_index), ][['split_index']],
      unified_model[!is.na(unified_model$Feature), ][['Node']]
    )
    expect_equal(
      to_test_features[['split_feature']],
      unified_model[['Feature']]
    )
    expect_equal(to_test_features[['threshold']], unified_model[['Split']])
    expect_equal(
      to_test_features[!is.na(internal_count), ][['internal_count']],
      unified_model[!is.na(unified_model$Feature), ][['Cover']]
    )
  })

  test_that('connections between nodes and leaves after lightgbm.unify are correct', {
    test_object <- as.data.table(lightgbm.unify(lgbm_fifa, sparse_data)$model)
    #Check if the sums of children's covers are correct
    expect_equal(
      test_object[test_object[!is.na(test_object$Yes)][['Yes']]][['Cover']] +
        test_object[test_object[!is.na(test_object$No)][['No']]][['Cover']],
      test_object[!is.na(Feature)][['Cover']]
    )
    #check if default_left information is correctly used
    df_default_left <- lgbmtree[
      default_left == "TRUE",
      c('tree_index', 'split_index')
    ]
    test_object_actual_default_left <- test_object[
      Yes == Missing,
      c('Tree', 'Node')
    ]
    colnames(test_object_actual_default_left) <- c('tree_index', 'split_index')
    attr(test_object_actual_default_left, 'model') <- NULL
    expect_equal(
      test_object_actual_default_left[order(tree_index, split_index)],
      df_default_left[order(tree_index, split_index)]
    )
    #and default_left = FALSE analogically:
    df_default_right <- lgbmtree[
      default_left != 'TRUE',
      c('tree_index', 'split_index')
    ]
    test_object_actual_default_right <- test_object[
      No == Missing,
      c('Tree', 'Node')
    ]
    colnames(test_object_actual_default_right) <- c('tree_index', 'split_index')
    attr(test_object_actual_default_right, 'model') <- NULL
    expect_equal(
      test_object_actual_default_right[order(tree_index, split_index)],
      df_default_right[order(tree_index, split_index)]
    )
    #One more test with checking the usage of 'decision_type' column needed
  })

  # Function that return the predictions for sample observations indicated by vector contatining values -1, 0, 1, where -1 means
  # going to the 'Yes' Node, 1 - to the 'No' node and 0 - to the missing node. The vectors are randomly produced during executing
  # the function and should be passed to prepare_original_preds_ to save the conscistence. Later we can compare the 'predicted' values
  prepare_test_preds <- function(unify_out) {
    stopifnot(all(
      c(
        "Tree",
        "Node",
        "Feature",
        "Split",
        "Yes",
        "No",
        "Missing",
        "Prediction",
        "Cover"
      ) %in%
        colnames(unify_out)
    ))
    test_tree <- unify_out[unify_out$Tree %in% 0:9, ]
    test_tree[['node_row_id']] <- seq_len(nrow(test_tree))
    test_obs <- lapply(table(test_tree$Tree), function(y) {
      sample(c(-1, 0, 1), y, replace = T)
    })
    test_tree <- split(test_tree, test_tree$Tree)
    determine_val <- function(obs, tree) {
      root_id <- tree[['node_row_id']][1]
      tree[, c('Yes', 'No', 'Missing')] <- tree[, c('Yes', 'No', 'Missing')] -
        root_id +
        1
      i <- 1
      indx <- 1
      while (!is.na(tree$Feature[indx])) {
        indx <- ifelse(
          obs[i] == 0,
          tree$Missing[indx],
          ifelse(obs[i] < 0, tree$Yes[indx], tree$No[indx])
        )
        #if(length(is.na(tree$Feature[indx]))>1) {print(paste(indx, i)); print(tree); print(obs)}
        i <- i + 1
      }
      return(tree[['Prediction']][indx])
    }
    x = numeric()
    for (i in seq_along(test_obs)) {
      x[i] <- determine_val(test_obs[[i]], test_tree[[i]])
    }
    return(list(preds = x, test_obs = test_obs))
  }

  prepare_original_preds_lgbm <- function(orig_tree, test_obs) {
    test_tree <- orig_tree[orig_tree$tree_index %in% 0:9, ]
    test_tree <- split(test_tree, test_tree$tree_index)
    stopifnot(length(test_tree) == length(test_obs))
    determine_val <- function(obs, tree) {
      i <- 1
      indx <- 1
      while (!is.na(tree$split_feature[indx])) {
        children <- ifelse(
          is.na(tree$node_parent),
          tree$leaf_parent,
          tree$node_parent
        )
        if ((obs[i] < 0) | (tree$default_left[indx] == 'TRUE' & obs[i] == 0)) {
          indx <- which(tree$split_index[indx] == children)[1]
        } else if (
          (obs[i] > 0) | (tree$default_left[indx] == 'FALSE' & obs[i] == 0)
        ) {
          indx <- which(tree$split_index[indx] == children)[2]
        } else {
          stop('Error in the connections')
          indx <- 0
        }
        i <- i + 1
      }
      return(tree[['leaf_value']][indx])
    }
    y = numeric()
    for (i in seq_along(test_obs)) {
      y[i] <- determine_val(test_obs[[i]], test_tree[[i]])
    }
    return(y)
  }

  test_that('the connections between the nodes are correct', {
    # The test is passed only if the predictions for sample observations are equal in the first 10 trees of the ensemble
    x <- prepare_test_preds(lightgbm.unify(lgbm_fifa, sparse_data)$model)
    preds <- x[['preds']]
    test_obs <- x[['test_obs']]
    original_preds <- prepare_original_preds_lgbm(lgbmtree, test_obs)
    expect_equal(preds, original_preds)
  })

  test_that("LightGBM: predictions from unified == original predictions", {
    unifier <- lightgbm.unify(lgbm_fifa, sparse_data)
    obs <- c(1:16000)
    original <- stats::predict(lgbm_fifa, sparse_data[obs, ])
    from_unified <- predict(unifier, sparse_data[obs, ])
    expect_equal(from_unified, original)
    #expect_true(all(abs((from_unified - original) / original) < 10**(-14))) #not needed
  })

  test_that("LightGBM: mean prediction calculated using predict == using covers", {
    unifier <- lightgbm.unify(lgbm_fifa, sparse_data)

    intercept_predict <- mean(predict(unifier, sparse_data))

    ntrees <- sum(unifier$model$Node == 0)
    leaves <- unifier$model[is.na(unifier$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) /
      sum(leaves$Cover) *
      ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })

  test_that("LightGBM: covers correctness", {
    unifier <- lightgbm.unify(lgbm_fifa, sparse_data)

    roots <- unifier$model[unifier$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(sparse_data)))

    internals <- unifier$model[!is.na(unifier$model$Feature), ]
    yes_child_cover <- unifier$model[internals$Yes, ]$Cover
    no_child_cover <- unifier$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unifier$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[
        internals$Missing == internals$Yes | internals$Missing == internals$No
      ] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })
}

if (requireNamespace("xgboost", quietly = TRUE)) {

  data <- fifa20$data[
    colnames(fifa20$data) != 'work_rate'
  ]
  target <- fifa20$target
  xgb_model <- xgboost::xgboost(
    x = as.matrix(data),
    y = target,
    nrounds = 100,
    objective = "reg:squarederror",
    max_depth = 3,
    nthreads = 1
  )
  xgb_tree <- xgboost::xgb.model.dt.tree(model = xgb_model)


  test_that('xgboost.unify returns an object of appropriate class', {
    expect_true(is.model_unified(xgboost.unify(xgb_model, as.matrix(data))))
    expect_true(is.model_unified(unify(xgb_model, as.matrix(data))))
  })

  test_that('xgboost.unify returns an object with correct attributes', {
    unified_model <- xgboost.unify(xgb_model, as.matrix(data))

    expect_equal(attr(unified_model, "missing_support"), TRUE)
    expect_equal(attr(unified_model, "model"), "xgboost")
  })

  test_that('columns after xgboost.unify are of appropriate type', {
    unified_model <- xgboost.unify(xgb_model, as.matrix(data))$model

    expect_true(is.integer(unified_model$Tree))
    expect_true(is.integer(unified_model$Node))
    expect_true(is.character(unified_model$Feature))
    expect_true(is.factor(unified_model$Decision.type))
    expect_true(is.numeric(unified_model$Split))
    expect_true(is.integer(unified_model$Yes))
    expect_true(is.integer(unified_model$No))
    expect_true(is.integer(unified_model$Missing))
    expect_true(is.numeric(unified_model$Prediction))
    expect_true(is.numeric(unified_model$Cover))
  })

  test_that('values in the columns after xgboost.unify are correct', {
    unified_model <- xgboost.unify(xgb_model, as.matrix(data))$model

    expect_equal(xgb_tree$Tree, unified_model$Tree)
    expect_equal(xgb_tree$Node, unified_model$Node)
    expect_equal(xgb_tree$Cover, unified_model$Cover)
    expect_equal(xgb_tree$Gain[xgb_tree$Feature == 'Leaf'], unified_model$Prediction[is.na(unified_model$Feature)])
    expect_equal(xgb_tree$Node, unified_model$Node)
    expect_equal(xgb_tree$Split, unified_model$Split)
    expect_equal(match(xgb_tree$Yes, xgb_tree$ID), unified_model$Yes)
    expect_equal(match(xgb_tree$No, xgb_tree$ID), unified_model$No)
    expect_equal(match(xgb_tree$Missing, xgb_tree$ID), unified_model$Missing)
    expect_equal(xgb_tree[xgb_tree[['Feature']] != 'Leaf',][['Feature']],
                unified_model[!is.na(unified_model$Feature),][['Feature']])
    expect_equal(nrow(xgb_tree[xgb_tree[['Feature']] == 'Leaf',]),
                nrow(unified_model[is.na(unified_model$Feature),]))

  })


  if (
    requireNamespace("lightgbm", quietly = TRUE) &&
      requireNamespace("jsonlite", quietly = TRUE)
  ) {
    test_that('xgboost.unify() does not work for objects produced with other packages', {
      param_lightgbm <- list(
        objective = "regression",
        max_depth = 2,
        num_leaves = 4L,
        force_row_wise = TRUE,
        learning.rate = 0.1
      )
      expect_warning({
        lgbm_fifa <- lightgbm::lightgbm(
          data = as.matrix(fifa20$data[colnames(fifa20$data) != 'value_eur']),
          label = fifa20$target,
          params = param_lightgbm,
          verbose = -1,
          num_threads = 0
        )
        lgbmtree <- lightgbm::lgb.model.dt.tree(lgbm_fifa)
      })
      expect_error(xgboost.unify(lgbmtree))
    })
  }

  # Function that return the predictions for sample observations indicated by vector contatining values -1, 0, 1, where -1 means
  # going to the 'Yes' Node, 1 - to the 'No' node and 0 - to the missing node. The vectors are randomly produced during executing
  # the function and should be passed to prepare_original_preds_ to save the conscistence. Later we can compare the 'predicted' values
  prepare_test_preds <- function(unify_out){
    stopifnot(all(c("Tree", "Node", "Feature", "Split", "Yes", "No", "Missing", "Prediction", "Cover") %in% colnames(unify_out)))
    test_tree <- unify_out[unify_out$Tree %in% 0:9, ]
    test_tree[['node_row_id']] <- seq_len(nrow(test_tree))
    test_obs <- lapply(table(test_tree$Tree), function(y) sample(c(-1, 0, 1), y, replace = T))
    test_tree <- split(test_tree, test_tree$Tree)
    determine_val <- function(obs, tree){
      root_id <- tree[['node_row_id']][1]
      tree[,c('Yes', 'No', 'Missing')] <- tree[,c('Yes', 'No', 'Missing')] - root_id + 1
      i <- 1
      indx <- 1
      while(!is.na(tree$Feature[indx])) {
        indx <- ifelse(obs[i] == 0, tree$Missing[indx], ifelse(obs[i] < 0, tree$Yes[indx], tree$No[indx]))
        i <- i + 1
      }
      return(tree[['Prediction']][indx])
    }
    x = numeric()
    for(i in seq_along(test_obs)) {
      x[i] <- determine_val(test_obs[[i]], test_tree[[i]])

    }
    return(list(preds = x, test_obs = test_obs))
  }

  prepare_original_preds_xgb <- function(orig_tree, test_obs){
    test_tree <- orig_tree[orig_tree$Tree %in% 0:9, ]
    test_tree <- split(test_tree, test_tree$Tree)
    stopifnot(length(test_tree) == length(test_obs))
    determine_val <- function(obs, tree){
      i <- 1
      indx <- 1
      while(!is.na(tree$Split[indx])) {
        indx <- ifelse(obs[i] == 0, match(tree$Missing[indx], tree$ID), ifelse(obs[i] < 0, match(tree$Yes[indx], tree$ID),
                                                                              match(tree$No[indx], tree$ID)))

        i <- i + 1
      }
      return(tree[['Gain']][indx])
    }
    y = numeric()
    for(i in seq_along(test_obs)) {
      y[i] <- determine_val(test_obs[[i]], test_tree[[i]])
    }
    return(y)
  }

  test_that('the connections between the nodes are correct', {
    # The test is passed only if the predictions for sample observations are equal in the first 10 trees of the ensemble
    x <- prepare_test_preds(xgboost.unify(xgb_model, as.matrix(data))$model)
    preds <- x[['preds']]
    test_obs <- x[['test_obs']]
    original_preds <- prepare_original_preds_xgb(xgb_tree, test_obs)
    expect_equal(preds, original_preds)
  })


  test_that("xgboost: covers correctness", {
    unifier <- xgboost.unify(xgb_model, data)

    roots <- unifier$model[unifier$model$Node == 0, ]
    expect_true(all(roots$Cover == nrow(data)))

    internals <- unifier$model[!is.na(unifier$model$Feature), ]
    yes_child_cover <- unifier$model[internals$Yes, ]$Cover
    no_child_cover <- unifier$model[internals$No, ]$Cover
    if (all(is.na(internals$Missing))) {
      children_cover <- yes_child_cover + no_child_cover
    } else {
      missing_child_cover <- unifier$model[internals$Missing, ]$Cover
      missing_child_cover[is.na(missing_child_cover)] <- 0
      missing_child_cover[
        internals$Missing == internals$Yes | internals$Missing == internals$No
      ] <- 0
      children_cover <- yes_child_cover + no_child_cover + missing_child_cover
    }
    expect_true(all(internals$Cover == children_cover))
  })

  # for the next tests, restore behaviour regarding output of 'xgboost::xgb.model.dt.tree'
  # from xgboost's api <v2.0 (intercept is now handeled differently,
  # see https://xgboost.readthedocs.io/en/release_3.1.0/tutorials/intercept.html and
  # https://xgboost.readthedocs.io/en/release_3.1.0/parameter.html#learning-task-parameters for further information on parameter 'base_score')
  xgb_model2 <- xgboost::xgboost(
    x = data,
    y = target,
    nrounds = 100,
    objective = "reg:squarederror",
    max_depth = 3,
    base_score = 0
  )

  unifier <- xgboost.unify(xgb_model2, data)
  # hack, to make this unit-test working
  unifier$model[
    which(unifier$model$Decision.type == "<="),
    "Decision.type"
  ] <- "<"

  test_that("xgboost: predictions from unified == original predictions", {
    obs <- data[1:16000, ]
    original <- stats::predict(xgb_model2, obs)
    from_unified <- predict(unifier, obs)
    # for debugging:
    #abs((from_unified - original) / original) |> max()
    #abs((from_unified - original) / original) |> mean()
    # expect_equal(from_unified, original) #there are small differences
    expect_true(all(abs((from_unified - original) / original) < 5 * 10**(-3)))
  })

  test_that("xgboost: mean prediction calculated using predict == using covers", {
    intercept_predict <- mean(predict(unifier, data))

    ntrees <- sum(unifier$model$Node == 0)
    leaves <- unifier$model[is.na(unifier$model$Feature), ]
    intercept_covers <- sum(leaves$Prediction * leaves$Cover) / sum(leaves$Cover) * ntrees

    #expect_true(all(abs((intercept_predict - intercept_covers) / intercept_predict) < 10**(-14)))
    expect_equal(intercept_predict, intercept_covers)
  })
}

# should be preceded with lgb.model.dt.tree
#' Unify LightGBM model
#'
#' Convert your LightGBM model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @param lgb_model A lightgbm model - object of class \code{lgb.Booster}
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#' @param recalculate logical indicating if covers should be recalculated according to the dataset given in data. Keep it \code{FALSE} if training data are used.
#'
#' @return a unified model representation - a \code{\link{model_unified.object}} object
#'
#' @export
#'
#' @seealso
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("lightgbm", quietly = TRUE) &&
#'  requireNamespace("jsonlite", quietly = TRUE)) {
#'   library(lightgbm)
#'   param_lgbm <- list(objective = "regression", max_depth = 2,
#'                      force_row_wise = TRUE, num_iterations = 20)
#'   data_fifa <- fifa20$data[!colnames(fifa20$data) %in%
#'                c('work_rate', 'value_eur', 'gk_diving', 'gk_handling',
#'                'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
#'   data <- na.omit(cbind(data_fifa, fifa20$target))
#'   sparse_data <- as.matrix(data[,-ncol(data)])
#'   x <- lightgbm::lgb.Dataset(sparse_data, label = as.matrix(data[,ncol(data)]))
#'   lgb_data <- lightgbm::lgb.Dataset.construct(x)
#'   lgb_model <- lightgbm::lightgbm(data = lgb_data, params = param_lgbm,
#'                                   verbose = -1, num_threads = 1)
#'   unified_model <- lightgbm.unify(lgb_model, sparse_data)
#'   shaps <- treeshap(unified_model, data[1:2, ])
#'   plot_contribution(shaps, obs = 1)
#' }}
lightgbm.unify <- function(lgb_model, data, recalculate = FALSE) {
  if (!requireNamespace("lightgbm", quietly = TRUE)) {
    stop("Package \"lightgbm\" needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop(
      "Package \"jsonlite\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }
  df <- lightgbm::lgb.model.dt.tree(lgb_model)
  stopifnot(c("split_index", "split_feature", "node_parent", "leaf_index", "leaf_parent", "internal_value",
              "internal_count", "leaf_value", "leaf_count", "decision_type") %in% colnames(df))
  df <- data.table::as.data.table(df)
  #convert node_parent and leaf_parent into one parent column
  df[is.na(df$node_parent), "node_parent"] <- df[is.na(df$node_parent), "leaf_parent"]
  #convert values into one column...
  df[is.na(df$internal_value), "internal_value"] <- df[!is.na(df$leaf_value), "leaf_value"]
  #...and counts
  df[is.na(df$internal_count), "internal_count"] <- df[!is.na(df$leaf_count), "leaf_count"]
  df[["internal_count"]] <- as.numeric(df[["internal_count"]])
  #convert split_index and leaf_index into one column:
  max_split_index <- df[, max(split_index, na.rm = TRUE), tree_index]
  rep_max_split <- rep(max_split_index$V1, times = as.numeric(table(df$tree_index)))
  new_leaf_index <- rep_max_split + df[, "leaf_index"] + 1
  df[is.na(df$split_index), "split_index"] <- new_leaf_index[!is.na(new_leaf_index[["leaf_index"]]), 'leaf_index']
  df[is.na(df$split_gain), "split_gain"] <- df[is.na(df$split_gain), "leaf_value"]
  # On the basis of column 'Parent', create columns with childs: 'Yes', 'No' and 'Missing' like in the xgboost df:
  ret.first <- function(x) x[1]
  ret.second <- function(x) x[2]
  tmp <- data.table::merge.data.table(
    df[, .(node_parent, tree_index, split_index)],
    df[, .(tree_index, split_index, default_left, decision_type)],
    by.x = c("tree_index", "node_parent"),
    by.y = c("tree_index", "split_index")
  )
  y_n_m <- unique(tmp[,
    .(
      Yes = ifelse(
        decision_type %in% c("<=", "<"),
        ret.first(split_index),
        ifelse(
          decision_type %in% c(">=", ">"),
          ret.second(split_index),
          stop("Unknown decision_type")
        )
      ),
      No = ifelse(
        decision_type %in% c(">=", ">"),
        ret.first(split_index),
        ifelse(
          decision_type %in% c("<=", "<"),
          ret.second(split_index),
          stop("Unknown decision_type")
        )
      ),
      Missing = ifelse(
        default_left,
        ret.first(split_index),
        ret.second(split_index)
      ),
      decision_type = decision_type
    ),
    .(tree_index, node_parent)
  ])
  df <- data.table::merge.data.table(
    df[, c(
      "tree_index",
      "depth",
      "split_index",
      "split_feature",
      "node_parent",
      "split_gain",
      "threshold",
      "internal_value",
      "internal_count"
    )],
    y_n_m,
    by.x = c("tree_index", "split_index"),
    by.y = c("tree_index", "node_parent"),
    all.x = TRUE
  )
  df[decision_type == ">=", decision_type := "<"]
  df[decision_type == ">", decision_type := "<="]
  df$Decision.type <- factor(x = df$decision_type, levels = c("<=", "<"))
  df[is.na(split_index), Decision.type := NA]
  df <- df[, c("tree_index", "split_index", "split_feature", "Decision.type", "threshold", "Yes", "No", "Missing", "split_gain", "internal_count")]
  colnames(df) <- c("Tree", "Node", "Feature", "Decision.type", "Split", "Yes", "No", "Missing", "Prediction", "Cover")
  attr(df, "sorted") <- NULL

  ID <- paste0(df$Node, "-", df$Tree)
  df$Yes <- match(paste0(df$Yes, "-", df$Tree), ID)
  df$No <- match(paste0(df$No, "-", df$Tree), ID)
  df$Missing <- match(paste0(df$Missing, "-", df$Tree), ID)

  # Here we lose "Quality" information
  df$Prediction[!is.na(df$Feature)] <- NA

  feature_names <- jsonlite::fromJSON(lgb_model$dump_model())$feature_names
  data <- data[,colnames(data) %in% feature_names]

  ret <- list(model = as.data.frame(df), data = as.data.frame(data), feature_names = feature_names)
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- TRUE
  attr(ret, "model") <- "LightGBM"

  if (recalculate) {
    ret <- set_reference_dataset(ret, as.data.frame(data))
  }

  return(ret)
}

#' Attributes of all players in FIFA 20
#'
#' Dataset consists of 56 columns, 55 numeric and one of type factor \code{'work_rate'}.
#' \code{value_eur} is a potential target feature.
#'
#' @format A data frame with 18278 rows and 56 columns.
#' Most of variables representing skills are in range from 0 to 100 and will not be described here.
#' To list non obvious features:
#' \describe{
#' \item{overall}{Overall score of player's skills}
#' \item{potential}{Potential of a player, younger players tend to have higher level of potential}
#' \item{value_eur}{Market value of a player (in mln EUR)}
#' \item{international_reputation}{Range 1 to 5}
#' \item{weak_foot}{Range 1 to 5}
#' \item{skill_moves}{Range 1 to 5}
#' \item{work_rate}{Divided by slash levels of willingness to work in offense and defense respectively}
#' }
#'
#'@source
#'"Data has been scraped from the publicly available website \url{https://sofifa.com}"
#'\url{https://www.kaggle.com/stefanoleone992/fifa-20-complete-player-dataset}
#'
"fifa20"

#' Unify ranger survival model
#'
#' Convert your ranger model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @details
#' The survival forest implemented in the \code{ranger} package stores cumulative hazard
#' functions (CHFs) in the leaves of survival trees, as proposed for Random Survival Forests
#' (Ishwaran et al. 2008). The final model prediction is made by averaging these CHFs
#' from all the trees. To provide explanations in the form of a survival function,
#' the CHFs from the leaves are converted into survival functions (SFs) using
#' the formula SF(t) = exp(-CHF(t)).
#' However, it is important to note that averaging these SFs does not yield the correct
#' model prediction as the model prediction is the average of CHFs transformed in the same way.
#' Therefore, when you obtain explanations based on the survival function,
#' they are only proxies and may not be fully consistent with the model predictions
#' obtained using for example \code{predict} function.
#'
#
#' @param rf_model An object of \code{ranger} class. At the moment, models built on data with categorical features
#' are not supported - please encode them before training.
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#' @param type A character to define the type of model prediction to use. Either `"risk"` (default), which uses the risk score calculated as a sum of cumulative hazard function values, `"survival"`, which uses the survival probability at certain time-points for each observation, or `"chf"`, which used the cumulative hazard values at certain time-points for each observation.
#' @param times A numeric vector of unique death times at which the prediction should be evaluated. By default `unique.death.times` from model are used.
#'
#' @return For `type = "risk"` a unified model representation is returned - a \code{\link{model_unified.object}} object. For `type = "survival"` or `type = "chf"` - a \code{\link{model_unified_multioutput.object}} object is returned, which is a list that contains unified model representation (\code{\link{model_unified.object}} object) for each time point. In this case, the list names are time points at which the survival function was evaluated.
#'
#' @export
#'
#' @seealso
#' \code{\link{ranger.unify}} for regression and classification \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' if (requireNamespace("ranger", quietly = TRUE) &&
#'  requireNamespace("survival", quietly = TRUE)) {
#'
#'   library(ranger)
#'   data_colon <- data.table::data.table(survival::colon)
#'   data_colon <- na.omit(data_colon[get("etype") == 2, ])
#'   surv_cols <- c("status", "time", "rx")
#'  
#'   feature_cols <- colnames(data_colon)[3:(ncol(data_colon) - 1)]
#'  
#'   train_x <- model.matrix(
#'     ~ -1 + .,
#'     data_colon[, .SD, .SDcols = setdiff(feature_cols, surv_cols[1:2])]
#'   )
#'   train_y <- survival::Surv(
#'     event = (data_colon[, get("status")] |>
#'                as.character() |>
#'                as.integer()),
#'     time = data_colon[, get("time")],
#'     type = "right"
#'   )
#'  
#'   rf <- ranger::ranger(
#'     x = train_x,
#'     y = train_y,
#'     data = data_colon,
#'     max.depth = 10,
#'     num.trees = 10
#'   )
#'   unified_model_risk <- ranger_surv.unify(rf, train_x, type = "risk")
#'   shaps <- treeshap(unified_model_risk, train_x[1:2,])
#'  
#'   # compute shaps for 3 selected time points
#'   unified_model_surv <- ranger_surv.unify(rf, train_x, type = "survival", times = c(23  , 50, 73))
#'   shaps_surv <- treeshap(unified_model_surv, train_x[1:2,])
#' }
#'
ranger_surv.unify <- function(rf_model, data, type = c("risk", "survival", "chf"), times = NULL) {
  type <- match.arg(type)

  stopifnot(
    "`times` must be a numeric vector and argument \
    `type = 'survival'` or `type = 'chf'` must be set." =
      ifelse(!is.null(times), is.numeric(times) && (type == "survival" || type == "chf"), TRUE)
  )

  surv_common <- ranger_surv.common(rf_model, data)
  n <- surv_common$n
  chf_table_list <- surv_common$chf_table_list

  if (type == "risk") {

    x <- lapply(chf_table_list, function(tree) {
      tree_data <- tree$tree_data
      nodes_chf <- tree$table
      tree_data$prediction <- rowSums(nodes_chf)
      tree_data[, c("nodeID", "leftChild", "rightChild", "splitvarName",
                    "splitval", "prediction")]
    })
    unified_return <- ranger_unify.common(x = x, n = n, data = data, feature_names = rf_model$forest$independent.variable.names)

  } else if (type == "survival" || type == "chf") {

    unique_death_times <- rf_model$unique.death.times

    if (is.null(times)) {
      compute_at_times <- unique_death_times
      # eval_times is required for list names (mainly when eval-times are
      # differing to the unique death times from the model as in the next case)
      eval_times <- as.character(compute_at_times)
    } else {
      stepfunction <- stepfun(unique_death_times, c(unique_death_times[1], unique_death_times))
      compute_at_times <- stepfunction(times)
      eval_times <- as.character(times)
    }

    # iterate over time-points
    unified_return <- lapply(compute_at_times, function(t) {
      time_index <- which(unique_death_times == t)
      x <- lapply(chf_table_list, function(tree) {
        tree_data <- tree$tree_data
        nodes_chf <- tree$table[, time_index]

        # transform cumulative hazards to survival function (if needed)
        # H(t) = -ln(S(t))
        # S(t) = exp(-H(t))
        tree_data$prediction <- if(type == "survival") exp(-nodes_chf) else nodes_chf
        tree_data[, c("nodeID", "leftChild", "rightChild", "splitvarName",
                      "splitval", "prediction")]
      })
      ranger_unify.common(x = x, n = n, data = data, feature_names = rf_model$forest$independent.variable.names)
    })
    names(unified_return) <- eval_times
    class(unified_return) <- "model_unified_multioutput"
  }
  return(unified_return)
}

ranger_surv.common <- function(rf_model, data) {
  if (!"ranger" %in% class(rf_model)) {
    stop("Object rf_model was not of class \"ranger\"")
  }
  if (!rf_model$treetype == "Survival") {
    stop("Object rf_model is not a random survival forest.")
  }
  n <- rf_model$num.trees
  chf_table_list <- lapply(1:n, function(tree) {
    tree_data <- data.table::as.data.table(ranger::treeInfo(rf_model,
                                                            tree = tree))

    # first get number of columns
    chf_node <- rf_model$forest$chf[[tree]]
    nodes_chf_n <- ncol(do.call(rbind, chf_node))
    nodes_prepare_chf_list <- lapply(
      X = chf_node,
      FUN = function(node) {
        if (identical(node, numeric(0L))) {
          rep(NA, nodes_chf_n)
        } else {
          node
        }
      }
    )
    list(table = do.call(rbind, nodes_prepare_chf_list), tree_data = tree_data)
  })
  return(list(chf_table_list = chf_table_list, n = n))
}

#' DrWhy Theme for ggplot objects
#'
#' @return theme for ggplot2 objects
#' @export
#' @rdname theme_drwhy
theme_drwhy <- function() {
  theme_bw(base_line_size = 0) %+replace%
    theme(
      axis.ticks = element_blank(),
      legend.background = element_blank(),
      legend.key = element_blank(),
      panel.background = element_blank(),
      panel.border = element_blank(),
      strip.background = element_blank(),
      plot.background = element_blank(),
      complete = TRUE,
      legend.direction = "horizontal",
      legend.position = "top",
      axis.line.y = element_line(color = "white"),
      axis.ticks.y = element_line(color = "white"),
      #axis.line = element_line(color = "#371ea3", linewidth = 0.5, linetype = 1),
      axis.title = element_text(color = "#371ea3"),
      plot.title = element_text(color = "#371ea3", size = 16, hjust = 0),
      plot.subtitle = element_text(color = "#371ea3", hjust = 0),
      axis.text = element_text(color = "#371ea3", size = 10),
      strip.text = element_text(color = "#371ea3", size = 12, hjust = 0),
      panel.grid.major.y = element_line(
        color = "grey90",
        linewidth = 0.5,
        linetype = 1
      ),
      panel.grid.minor.y = element_line(
        color = "grey90",
        linewidth = 0.5,
        linetype = 1
      ),
      panel.grid.minor.x = element_blank(),
      panel.grid.major.x = element_blank()
    )

}

#' @export
#' @rdname theme_drwhy
theme_drwhy_vertical <- function() {
  theme_bw(base_line_size = 0) %+replace%
    theme(
      axis.ticks = element_blank(),
      legend.background = element_blank(),
      legend.key = element_blank(),
      panel.background = element_blank(),
      panel.border = element_blank(),
      strip.background = element_blank(),
      plot.background = element_blank(),
      complete = TRUE,
      legend.direction = "horizontal",
      legend.position = "top",
      axis.line.x = element_line(color = "white"),
      axis.ticks.x = element_line(color = "white"),
      plot.title = element_text(color = "#371ea3", size = 16, hjust = 0),
      plot.subtitle = element_text(color = "#371ea3", hjust = 0),
      #axis.line = element_line(color = "#371ea3", linewidth = 0.5, linetype = 1),
      axis.title = element_text(color = "#371ea3"),
      axis.text = element_text(color = "#371ea3", size = 10),
      strip.text = element_text(color = "#371ea3", size = 12, hjust = 0),
      panel.grid.major.x = element_line(
        color = "grey90",
        linewidth = 0.5,
        linetype = 1
      ),
      panel.grid.minor.x = element_line(
        color = "grey90",
        linewidth = 0.5,
        linetype = 1
      ),
      panel.grid.minor.y = element_blank(),
      panel.grid.major.y = element_blank()
    )
}

#' DrWhy color palettes for ggplot objects
#'
#' @param n number of colors for color palette
#'
#' @return color palette as vector of characters
#' @export
#' @rdname colors_drwhy
colors_discrete_drwhy <- function(n = 2) {
  if (n == 1) return("#4378bf")
  if (n == 2) return(c( "#4378bf", "#8bdcbe"))
  if (n == 3) return(c( "#4378bf", "#f05a71", "#8bdcbe"))
  if (n == 4) return(c( "#4378bf", "#f05a71", "#8bdcbe", "#ffa58c"))
  if (n == 5) return(c( "#4378bf", "#f05a71", "#8bdcbe", "#ae2c87", "#ffa58c"))
  if (n == 6) return(c( "#4378bf", "#46bac2", "#8bdcbe", "#ae2c87", "#ffa58c", "#f05a71"))
  c( "#4378bf", "#46bac2", "#371ea3", "#8bdcbe", "#ae2c87", "#ffa58c", "#f05a71")[((0:(n-1)) %% 7) + 1]
}

#' @export
#' @rdname colors_drwhy
colors_breakdown_drwhy <- function() {
  c(`-1` = "#f05a71", `0` = "#371ea3", `1` = "#8bdcbe", X = "#371ea3")
}

#' Unify ranger model
#'
#' Convert your ranger model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @param rf_model An object of \code{ranger} class. At the moment, models built on data with categorical features
#' are not supported - please encode them before training.
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#'
#' @return a unified model representation - a \code{\link{model_unified.object}} object
#'
#' @export
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' if (requireNamespace("ranger", quietly = TRUE)) {
#'  library(ranger)
#'  data_fifa <- fifa20$data[!colnames(fifa20$data) %in%
#'                             c('work_rate', 'value_eur', 'gk_diving', 'gk_handling',
#'                              'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
#'  data <- na.omit(cbind(data_fifa, target = fifa20$target))
#'
#'  rf <- ranger::ranger(target~., data = data, max.depth = 10, num.trees = 10)
#'  unified_model <- ranger.unify(rf, data)
#'  shaps <- treeshap(unified_model, data[1:2,])
#'  plot_contribution(shaps, obs = 1)
#' }
ranger.unify <- function(rf_model, data) {
  if(!'ranger' %in% class(rf_model)) {
    stop('Object rf_model was not of class "ranger"')
  }
  n <- rf_model$num.trees
  x <- lapply(1:n, function(tree) {
    tree_data <- data.table::as.data.table(ranger::treeInfo(rf_model, tree = tree))
    # Fix for probability forests
    if (rf_model$treetype == "Probability estimation") {
      data.table::setnames(tree_data, "pred.1", "prediction")
    }
    tree_data[, c("nodeID",  "leftChild", "rightChild", "splitvarName", "splitval", "prediction")]
  })
  return(ranger_unify.common(x = x, n = n, data = data, feature_names = rf_model$forest$independent.variable.names))
}


ranger_unify.common <- function(x, n, data, feature_names) {
  times_vec <- sapply(x, nrow)
  y <- data.table::rbindlist(x)
  y[, ("Tree") := rep(0:(n - 1), times = times_vec)]
  data.table::setnames(y, c("Node", "Yes", "No", "Feature", "Split",  "Prediction", "Tree"))
  y[, ("Feature") := as.character(get("Feature"))]
  y[y$Yes < 0, "Yes"] <- NA
  y[y$No < 0, "No"] <- NA
  y[, ("Missing") := NA]
  y$Cover <- 0
  y$Decision.type <- factor(x = rep("<=", times = nrow(y)), levels = c("<=", "<"))
  y[is.na(get("Feature")), ("Decision.type") := NA]

  ID <- paste0(y$Node, "-", y$Tree)
  y$Yes <- match(paste0(y$Yes, "-", y$Tree), ID)
  y$No <- match(paste0(y$No, "-", y$Tree), ID)

  # Here we lose "Quality" information
  y[!is.na(get("Feature")), ("Prediction") := NA]

  # treeSHAP assumes, that [prediction = sum of predictions of the trees]
  # in random forest [prediction = mean of predictions of the trees]
  # so here we correct it by adjusting leaf prediction values
  y[is.na(get("Feature")), ("Prediction") := I(get("Prediction") / n)]


  data.table::setcolorder(
    y, c("Tree", "Node", "Feature", "Decision.type", "Split",
         "Yes", "No", "Missing", "Prediction", "Cover"))

  data <- data[,colnames(data) %in% feature_names]

  ret <- list(model = as.data.frame(y), data = as.data.frame(data), feature_names = feature_names)
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- FALSE
  attr(ret, "model") <- "ranger"
  return(set_reference_dataset(ret, as.data.frame(data)))
}

#' Unify tree-based model
#'
#' Convert your tree-based model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @param model A tree-based model object of any supported class (\code{gbm}, \code{lgb.Booster}, \code{randomForest}, \code{ranger}, or \code{xgb.Booster}).
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#' @param ... Additional parameters passed to the model-specific unification functions.
#'
#' @return A unified model representation - a \code{\link{model_unified.object}} object (for single-output models) or \code{\link{model_unified_multioutput.object}}, which is a list of \code{\link{model_unified.object}} objects (for multi-output models).
#'
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gpboost.unify}} for \code{\link[gpboost:gpboost]{GPBoost models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @export
#'
#' @examples
#' if (
#'   requireNamespace("ranger", quietly = TRUE) &&
#'     requireNamespace("randomForest", quietly = TRUE)
#' ) {
#'   library(ranger)
#'   data_fifa <- fifa20$data[
#'     !colnames(fifa20$data) %in%
#'       c(
#'         'work_rate',
#'         'value_eur',
#'         'gk_diving',
#'         'gk_handling',
#'         'gk_kicking',
#'         'gk_reflexes',
#'         'gk_speed',
#'         'gk_positioning'
#'       )
#'   ]
#'   data <- na.omit(cbind(data_fifa, target = fifa20$target)) 
#'
#'   rf1 <- ranger::ranger(target ~ ., data = data, max.depth = 10, num.trees = 10)
#'   unified_model1 <- unify(rf1, data)
#'   shaps1 <- treeshap(unified_model1, data[1:2, ])
#'   plot_contribution(shaps1, obs = 1) 
#'
#'   rf2 <- randomForest::randomForest(
#'     target ~ .,
#'     data = data,
#'     maxnodes = 10,
#'     ntree = 10
#'   )
#'   unified_model2 <- unify(rf2, data)
#'   shaps2 <- treeshap(unified_model2, data[1:2, ])
#'   plot_contribution(shaps2, obs = 1)
#' }
#'
unify <- function(model, data, ...){
  UseMethod("unify", model)
}

#' @export
unify.gbm <- function(model, data, ...){
  gbm.unify(model, data)
}

#' @export
unify.lgb.Booster <- function(model, data, recalculate = FALSE, ...){
  lightgbm.unify(model, data, recalculate)
}

#' @export
unify.gpb.Booster <- function(model, data, recalculate = FALSE, ...){
  gpboost.unify(model, data, recalculate)
}

#' @export
unify.randomForest <- function(model, data, ...){
  randomForest.unify(model, data)
}

#' @export
unify.ranger <- function(model, data, ...){
  if (model$treetype == "Survival"){
    return(ranger_surv.unify(model, data, ...))
  }
  ranger.unify(model, data)
}

#' @export
unify.xgb.Booster <- function(model, data, recalculate = FALSE, ...){
  xgboost.unify(model, data, recalculate)
}

#' @export
unify.default <- function(model, data, ...){
  stop("Provided model is not of type supported by treeshap.")
}

#' SHAP value based Break-Down plot
#'
#' This function plots contributions of features into the prediction for a single observation.
#'
#' @param treeshap A treeshap object produced with the \code{\link{treeshap}} function. \code{\link{treeshap.object}}.
#' @param obs A numeric indicating which observation should be plotted. Be default it's first observation.
#' @param max_vars maximum number of variables that shall be presented. Variables with the highest importance will be presented.
#' Remaining variables will be summed into one additional contribution. By default \code{5}.
#' @param min_max a range of OX axis. By default \code{NA}, therefore it will be extracted from the contributions of \code{x}.
#' But it can be set to some constants, useful if these plots are to be used for comparisons.
#' @param digits number of decimal places (\code{\link{round}}) to be used.
#' @param explain_deviation if \code{TRUE} then instead of explaining prediction and plotting intercept bar, only deviation from mean prediction of the reference dataset will be explained. By default \code{FALSE}.
#' @param title the plot's title, by default \code{'SHAP Break-Down'}.
#' @param subtitle the plot's subtitle. By default no subtitle.
#'
#' @return a \code{ggplot2} object
#'
#' @export
#'
#' @seealso
#' \code{\link{treeshap}} for calculation of SHAP values
#'
#' \code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}, \code{\link{plot_interaction}}
#'
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE) &&
#'  requireNamespace("scales", quietly = TRUE)) {
#'   library(xgboost)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'   xgb_model <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 3,
#'    nrounds = 20,
#'    nthreads = 1
#'   )
#'   unified_model <- xgboost.unify(xgb_model, as.matrix(data))
#'   x <- head(data, 1)
#'   shap <- treeshap(unified_model, x)
#'   plot_contribution(shap, 1, min_max = c(-5e5, 1.3e8))
#' }}
#'
plot_contribution <- function(treeshap,
                              obs = 1,
                              max_vars = 5,
                              min_max = NA,
                              digits = 3,
                              explain_deviation = FALSE,
                              title = "SHAP Break-Down",
                              subtitle = "") {
  if (!requireNamespace("scales", quietly = TRUE)) {
    stop(
      "Package \"scales\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }

  shap <- treeshap$shaps[obs, ]
  model <- treeshap$unified_model$model
  x <- treeshap$observations[obs, ]

  # argument check
  if (!is.treeshap(treeshap)) {
    stop("treeshap parameter has to be correct object of class treeshap. Produce it using treeshap function.")
  }

  if (max_vars > ncol(shap)) {
    warning("max_vars exceeds number of variables. All variables will be shown.")
    max_vars <- ncol(shap)
  }
  if (nrow(shap) != 1) {
    warning("Only 1 observation can be plotted. Plotting 1st one.")
    shap <- shap[1, ]
  }

  # setting intercept
  mean_prediction <- mean(predict.model_unified(treeshap$unified_model, treeshap$unified_model$data))
  if (explain_deviation) {
    mean_prediction <- 0
  }

  df <- data.frame(variable = colnames(shap), contribution = as.numeric(shap))

  # setting variable names to showing their value
  df$variable <- paste0(df$variable, " = ", as.character(x))

  # selecting max_vars most important variables
  is_important <- order(abs(df$contribution), decreasing = TRUE)[1:max_vars]
  other_variables_contribution_sum <- sum(df$contribution[-is_important])
  df <- df[is_important, ]
  df$position <- 2:(max_vars + 1)
  if (max_vars < ncol(shap)) {
    df <- rbind(df, data.frame(variable = "+ all other variables",
                               contribution = other_variables_contribution_sum,
                               position = max(df$position) + 1))
  }

  # adding "prediction" bar
  df <- rbind(df, data.frame(variable = ifelse(explain_deviation, "prediction deviation", "prediction"),
                             contribution = mean_prediction + sum(df$contribution),
                             position = max(df$position) + 1))

  df$sign <- ifelse(df$contribution >= 0, "1", "-1")

  # adding "intercept" bar
  df <- rbind(df, data.frame(variable = "intercept",
                             contribution = mean_prediction,
                             position = 1,
                             sign = "X"))

  # ordering
  df <- df[order(df$position), ]

  # adding columns needed by plot
  df$cumulative <- cumsum(df$contribution)
  df$prev <- df$cumulative - df$contribution
  df$text <- as.character(round(df$contribution, digits))
  df$text[df$contribution > 0] <- paste0("+", df$text[df$contribution > 0])

  # intercept bar corrections:
  df$prev[1] <- df$contribution[1]
  df$text[1] <- as.character(round(df$contribution[1], digits))

  # prediction bar corrections:
  df$prev[nrow(df)] <- df$contribution[1]
  df$cumulative[nrow(df)] <- df$cumulative[max_vars + 2]
  if (!explain_deviation) { #  assuring it doesn't differ from prediction because of some numeric errors
    df$cumulative[nrow(df)] <- predict.model_unified(treeshap$unified_model, x)
  }
  df$sign[nrow(df)] <- "X"
  df$text[nrow(df)] <- as.character(round(df$contribution[nrow(df)], digits))

  # removing intercept bar if requested by explain_deviation argument
  if (explain_deviation) {
    df <- df[-1, ]
  }

  # reversing postions to sort bars decreasing
  df$position <- rev(df$position)

  # base plot
  p <- ggplot(df, aes(x = position + 0.5,
                      y = pmax(cumulative, prev),
                      xmin = position + 0.15, xmax = position + 0.85,
                      ymin = cumulative, ymax = prev,
                      fill = sign,
                      label = text))

  # add rectangles and hline
  p <- p +
    geom_errorbarh(data = df[-c(nrow(df), if (explain_deviation) nrow(df) - 1), ],
                   aes(xmax = position - 0.85,
                       xmin = position + 0.85,
                       y = cumulative), width = 0,
                   color = "#371ea3") +
    geom_rect(alpha = 0.9) +
    if (!explain_deviation) (geom_hline(data = df[df$variable == "intercept", ],
                                        aes(yintercept = contribution),
                                        lty = 3, alpha = 0.5, color = "#371ea3"))


  # add adnotations
  drange <- diff(range(df$cumulative))
  p <- p + geom_text(aes(y = pmax(cumulative,  cumulative - contribution)),
                     vjust = 0.5,
                     nudge_y = drange * 0.05,
                     hjust = 0,
                     color = "#371ea3")

  # set limits for contributions
  if (any(is.na(min_max))) {
    x_limits <- scale_y_continuous(expand = c(0.05, 0.15), name = "", labels = scales::comma)
  } else {
    x_limits <- scale_y_continuous(expand = c(0.05, 0.15), name = "", limits = min_max, labels = scales::comma)
  }

  p <- p + x_limits +
    scale_x_continuous(labels = df$variable, breaks = df$position + 0.5, name = "") +
    scale_fill_manual(values = colors_breakdown_drwhy())

  # add theme
  p + coord_flip() + theme_drwhy_vertical() +
    theme(legend.position = "none") +
    labs(title = title, subtitle = subtitle)
}

#' Unify XGBoost model
#'
#' Convert your XGBoost model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @param xgb_model A XGBoost model - object of class \code{xgb.Booster}
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#' @param recalculate logical indicating if covers should be recalculated according to the dataset given in data. Keep it \code{FALSE} if training data are used.
#'
#' @return a unified model representation - a \code{\link{model_unified.object}} object
#'
#' @export
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   library(xgboost)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'   xgb_model <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 3,
#'    nrounds = 20,
#'    nthreads = 1
#'   )
#'   unified_model <- xgboost.unify(xgb_model, as.matrix(data))
#'   shaps <- treeshap(unified_model, data[1:2,])
#'   plot_contribution(shaps, obs = 1)
#' }}
#'
xgboost.unify <- function(xgb_model, data, recalculate = FALSE) {
  if (!requireNamespace("xgboost", quietly = TRUE)) {
    stop(
      "Package \"xgboost\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }
  xgbtree <- xgboost::xgb.model.dt.tree(
    model = xgb_model
  )
  stopifnot(
    c(
      "Tree",
      "Node",
      "ID",
      "Feature",
      "Split",
      "Yes",
      "No",
      "Missing",
      "Gain",
      "Cover"
    ) %in%
      colnames(xgbtree)
  )
  xgbtree$Yes <- match(xgbtree$Yes, xgbtree$ID)
  xgbtree$No <- match(xgbtree$No, xgbtree$ID)
  xgbtree$Missing <- match(xgbtree$Missing, xgbtree$ID)

  xgbtree[is.na(xgbtree$Split), 'Feature'] <- NA_character_

  # xgboost split condition should be always "less than"
  # (https://xgboost.readthedocs.io/en/stable/r_docs/R-package/docs/reference/xgb.model.dt.tree.html#value)
  # however, when using "<" instead of "<=", xgboost related unit-tests from
  # 'test_treeshap_correctness.R' are failing.
  xgbtree$Decision.type <- factor(
    x = rep("<=", times = nrow(xgbtree)),
    levels = c("<=", "<")
  )
  xgbtree$Decision.type[is.na(xgbtree$Feature)] <- NA_character_
  xgbtree <- xgbtree[, c(
    "Tree",
    "Node",
    "Feature",
    "Decision.type",
    "Split",
    "Yes",
    "No",
    "Missing",
    "Gain",
    "Cover"
  )]
  colnames(xgbtree) <- c(
    "Tree",
    "Node",
    "Feature",
    "Decision.type",
    "Split",
    "Yes",
    "No",
    "Missing",
    "Prediction",
    "Cover"
  )

  # Here we lose "Quality" information
  xgbtree$Prediction[!is.na(xgbtree$Feature)] <- NA_real_

  feature_names <- xgboost::getinfo(xgb_model, "feature_name")
  data <- data[, colnames(data) %in% feature_names]

  ret <- list(
    model = as.data.frame(xgbtree),
    data = as.data.frame(data),
    feature_names = feature_names
  )
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- TRUE
  attr(ret, "model") <- "xgboost"

  if (recalculate) {
    ret <- set_reference_dataset(ret, as.data.frame(data))
  }

  return(ret)
}

if (requireNamespace("xgboost", quietly = TRUE)) {

  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target

  test_that('recalculate covers works correctly for xgboost model', {
    xgb_model <- xgboost::xgboost(
      x = as.matrix(data),
      y = target,
      objective = "reg:squarederror",
      max_depth = 5,
      nrounds = 100,
      nthreads = 1
    )
    unified <- xgboost.unify(xgb_model, as.matrix(data))
    a <- set_reference_dataset(unified, data)$Cover
    b <- unified$Cover
    expect_true(all(a == b))
  })
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

predict_cpp <- function(x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type, value) {
    .Call('_treeshap_predict_cpp', PACKAGE = 'treeshap', x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type, value)
}

new_covers <- function(x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type) {
    .Call('_treeshap_new_covers', PACKAGE = 'treeshap', x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type)
}

treeshap_cpp <- function(x, is_na, roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover, verbose) {
    .Call('_treeshap_treeshap_cpp', PACKAGE = 'treeshap', x, is_na, roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover, verbose)
}

treeshap_interactions_cpp <- function(x, is_na, roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover, verbose) {
    .Call('_treeshap_treeshap_interactions_cpp', PACKAGE = 'treeshap', x, is_na, roots, yes, no, missing, feature, split, decision_type, is_leaf, value, cover, verbose)
}

#' Set reference dataset
#'
#' Change a dataset used as reference for calculating SHAP values.
#' Reference dataset is initially set with \code{data} argument in unifying function.
#' Usually reference dataset is dataset used to train the model.
#' Important property of reference dataset is that SHAP values for each observation add up to its deviation from mean prediction for a reference dataset.
#'
#'
#' @param unified_model Unified model representation of the model created with a (model).unify function. (\code{\link{model_unified.object}}).
#' @param x Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model.
#'
#' @return  \code{\link{model_unified.object}}. Unified representation of the model as created with a (model).unify function,
#' but with changed reference dataset (Cover column containing updated values).
#'
#' @export
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("gbm", quietly = TRUE)) {
#'   library(gbm)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   data['value_eur'] <- fifa20$target
#'   gbm_model <- gbm::gbm(
#'   formula = value_eur ~ .,
#'     data = data,
#'     distribution = "laplace",
#'     n.trees = 20,
#'     interaction.depth = 4,
#'     n.cores = 1)
#'   unified <- gbm.unify(gbm_model, data)
#'   set_reference_dataset(unified, data[200:700, ])
#' }}
set_reference_dataset <- function(unified_model, x) {
  model <- unified_model$model
  data <- x

  # argument check
  if (!("matrix" %in% class(x) | "data.frame" %in% class(x))) {
    stop("x parameter has to be data.frame or matrix.")
  }

  if (!("model_unified" %in% class(unified_model))) {
    stop("unified_model parameter has to of class model_unified. Produce it using *.unify function.")
  }

  if (!all(c("Tree", "Node", "Feature", "Decision.type", "Split", "Yes", "No", "Missing", "Prediction") %in% colnames(model))) {
    stop("Given model dataframe is not a correct unified dataframe representation. Use (model).unify function.")
  }

  if (!attr(unified_model, "missing_support") && any(is.na(x))) {
    stop("Given model does not work with missing values. Dataset x should not contain missing values.")
  }

  x <- x[,colnames(x) %in% unified_model$feature_names]

  if (!all(model$Feature %in% c(NA, colnames(x)))) {
    stop("Dataset does not contain all features occurring in the model.")
  }


  # adapting model representation to C++ and extracting from dataframe to vectors
  roots <- which(model$Node == 0) - 1
  yes <- model$Yes - 1
  no <- model$No - 1
  missing <- model$Missing - 1
  is_leaf <- is.na(model$Feature)
  feature <- match(model$Feature, colnames(x)) - 1
  split <- model$Split
  decision_type <- unclass(model$Decision.type)
  #stopifnot(levels(decision_type) == c("<=", "<"))
  #stopifnot(all(decision_type %in% c(1, 2, NA)))

  n <- nrow(x)
  x <- as.data.frame(sapply(x, as.numeric))
  if (n > 1) x <- t(x)
  is_na <- is.na(x) # needed, because dataframe passed to cpp somehow replaces missing values with random values

  model$Cover <- new_covers(x, is_na, roots, yes, no, missing, is_leaf, feature, split, decision_type)

  ret <- list(model = as.data.frame(model), data = as.data.frame(data), feature_names = unified_model$feature_names)
  #attributes(ret) <- attributes(model_unified)
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- attr(unified_model, "missing_support")
  attr(ret, 'model') <- attr(unified_model, "model")

  return(ret)
}

#' Unify GBM model
#'
#' Convert your GBM model into a standardized representation.
#' The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
#'
#' @param gbm_model An object of \code{gbm} class. At the moment, models built on data with categorical features
#' are not supported - please encode them before training.
#' @param data Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.
#'
#' @return a unified model representation - a \code{\link{model_unified.object}} object
#'
#' @export
#'
#' @seealso
#' \code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}
#'
#' \code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}
#'
#' \code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
#'
#' \code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("gbm", quietly = TRUE)) {
#'   library(gbm)
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   data['value_eur'] <- fifa20$target
#'   gbm_model <- gbm::gbm(
#'                formula = value_eur ~ .,
#'                data = data,
#'                distribution = "gaussian",
#'                n.trees = 20,
#'                interaction.depth = 4,
#'                n.cores = 1)
#'   unified_model <- gbm.unify(gbm_model, data)
#'   shaps <- treeshap(unified_model, data[1:2,])
#'   plot_contribution(shaps, obs = 1)
#' }}
gbm.unify <- function(gbm_model, data) {
  if(!inherits(gbm_model,'gbm')) {
    stop('Object gbm_model was not of class "gbm"')
  }
  if(any(gbm_model$var.type > 0)) {
    stop('Models built on data with categorical features are not supported - please encode them before training.')
  }
  x <- lapply(gbm_model$trees, data.table::as.data.table)
  times_vec <- sapply(x, nrow)
  y <- data.table::rbindlist(x)
  data.table::setnames(y, c("Feature", "Split", "Yes",
                            "No", "Missing", "ErrorReduction", "Cover",
                            "Prediction"))
  y[["Tree"]] <- rep(0:(length(gbm_model$trees) - 1), times = times_vec)
  y[["Node"]] <- unlist(lapply(times_vec, function(x) 0:(x - 1)))
  y <- y[, Feature := as.character(Feature)]
  y[y$Feature < 0, "Feature"] <- NA
  y[!is.na(y$Feature), "Feature"] <- attr(gbm_model$Terms, "term.labels")[as.integer(y[["Feature"]][!is.na(y$Feature)]) + 1]
  y[is.na(y$Feature), "ErrorReduction"] <- y[is.na(y$Feature), "Split"]
  y[is.na(y$Feature), "Split"] <- NA
  y[y$Yes < 0, "Yes"] <- NA
  y[y$No < 0, "No"] <- NA
  y[y$Missing < 0, "Missing"] <- NA
  y$Decision.type <- factor(x = rep("<=", times = nrow(y)), levels = c("<=", "<"))
  y[is.na(Feature), Decision.type := NA]
  y <- y[, c("Tree", "Node", "Feature", "Decision.type", "Split", "Yes", "No", "Missing", "ErrorReduction", "Cover")]
  colnames(y) <- c("Tree", "Node", "Feature", "Decision.type", "Split", "Yes", "No", "Missing", "Prediction", "Cover")

  ID <- paste0(y$Node, "-", y$Tree)
  y$Yes <- match(paste0(y$Yes, "-", y$Tree), ID)
  y$No <- match(paste0(y$No, "-", y$Tree), ID)
  y$Missing <- match(paste0(y$Missing, "-", y$Tree), ID)

  # Here we lose "Quality" information
  y[!is.na(Feature), Prediction := NA]

  # GBM calculates prediction as [initF + sum of predictions of trees]
  # treeSHAP assumes prediction are calculated as [sum of predictions of trees]
  # so here we adjust it
  ntrees <- sum(y$Node == 0)
  y[is.na(Feature), Prediction := Prediction + gbm_model$initF / ntrees]

  feature_names <- gbm_model$var.names
  data <- data[,colnames(data) %in% feature_names]

  ret <- list(model = as.data.frame(y), data = as.data.frame(data), feature_names = feature_names)
  class(ret) <- "model_unified"
  attr(ret, "missing_support") <- TRUE
  attr(ret, "model") <- "gbm"

  # Original covers in gbm_model are not correct
  ret <- set_reference_dataset(ret, as.data.frame(data))

  return(ret)
}

#' SHAP Interaction value plot
#'
#' This function plots SHAP Interaction value for two variables depending on the value of the first variable.
#' Value of the second variable is marked with the color.
#'
#' @param treeshap A treeshap object produced with \code{\link{treeshap}(interactions = TRUE)} function. \code{\link{treeshap.object}}.
#' @param var1 name or index of the first variable - plotted on x axis.
#' @param var2 name or index of the second variable - marked with color.
#' @param title the plot's title, by default \code{'SHAP Interaction Value Plot'}.
#' @param subtitle the plot's subtitle. By default no subtitle.
#'
#' @return a \code{ggplot2} object
#'
#' @export
#'
#' @seealso
#' \code{\link{treeshap}} for calculation of SHAP Interaction values
#'
#' \code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}
#'
#'
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE) &&
#'  requireNamespace("scales", quietly = TRUE)) {
#'   data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
#'   target <- fifa20$target
#'   xgb_model2 <- xgboost::xgboost(
#'    x = as.matrix(data),
#'    y = target,
#'    objective = "reg:squarederror",
#'    max_depth = 5,
#'    nrounds = 10,
#'    nthreads = 1
#'   )
#'   unified_model2 <- xgboost.unify(xgb_model2, data)
#'   inters <- treeshap(unified_model2, as.matrix(data[1:50, ]), interactions = TRUE)
#'   plot_interaction(inters, "dribbling", "defending")
#' }}
plot_interaction <- function(treeshap, var1, var2,
                             title = "SHAP Interaction Value Plot",
                             subtitle = "") {
  if (!requireNamespace("scales", quietly = TRUE)) {
    stop(
      "Package \"scales\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }

  interactions <- treeshap$interactions
  x <- treeshap$observations

  # argument check
  if (!is.treeshap(treeshap)) {
    stop("treeshap parameter has to be correct object of class treeshap. Produce it using treeshap function.")
  }

  if (is.null(interactions)) {
    stop("SHAP Interaction values were not calculated in treeshap object. You need to use treeshap(interactions = TRUE).")
  }

  if (is.character(var1)) {
    if (!(var1 %in% colnames(x))) stop("var1 is not a correct variable name. It does not occur in the dataset.")
    if (!(var1 %in% colnames(interactions))) stop("var1 is not a correct variable name. It does not occur in interactions object.")
  } else if (is.numeric(var1)) {
    if (var1 > ncol(x) || var1 < 1) stop("var1 is not a correct number.")
  }

  if (is.character(var2)) {
    if (!(var2 %in% colnames(x))) stop("var2 is not a correct variable name. It does not occur in the dataset.")
    if (!(var2 %in% colnames(interactions))) stop("var2 is not a correct variable name. It does not occur in interactions object.")
  } else if (is.numeric(var2)) {
    if (var2 > ncol(x) || var2 < 1) stop("var2 is not a correct number.")
  }


  interaction <- interactions[var1, var2, ]
  var1_value <- x[,var1]
  var2_value <- x[,var2]
  plot_data <- data.frame(var1_value = var1_value, var2_value = var2_value, interaction = interaction)

  x_lab <- ifelse(is.character(var1), var1, colnames(x)[var1])
  col_lab <- ifelse(is.character(var2), var2, colnames(x)[var2])
  y_lab <- paste0("SHAP Interaction value for ", x_lab, " and ", col_lab)

  p <- ggplot(plot_data, aes(x = var1_value, y = interaction, color = var2_value)) +
    geom_point() +
    labs(x = x_lab, color = col_lab, y = y_lab, title = title, subtitle = subtitle) +
    scale_y_continuous(labels = scales::comma) +
    theme_drwhy()
  p
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{print.treeshap}
\alias{print.treeshap}
\title{Prints treeshap objects}
\usage{
\method{print}{treeshap}(x, ...)
}
\arguments{
\item{x}{a treeshap object}

\item{...}{other arguments}
}
\value{
No return value, called for printing
}
\description{
Prints treeshap objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_unified.R
\name{model_unified.object}
\alias{model_unified.object}
\title{Unified model representation}
\value{
List consisting of two elements:

\strong{model} - A \code{data.frame} representing model with following columns:

\item{Tree}{0-indexed ID of a tree}
\item{Node}{0-indexed ID of a node in a tree. In a tree the root always has ID 0}
\item{Feature}{In case of an internal node - name of a feature to split on. Otherwise - NA}
\item{Decision.type}{A factor with two levels: "<" and "<=". In case of an internal node - predicate used for splitting observations. Otherwise - NA}
\item{Split}{For internal nodes threshold used for splitting observations. All observations that satisfy the predicate Decision.type(Split) ('< Split' / '<= Split') are proceeded to the node marked as 'Yes'. Otherwise to the 'No' node. For leaves - NA}
\item{Yes}{Index of a row containing a child Node. Thanks to explicit indicating the row it is much faster to move between nodes}
\item{No}{Index of a row containing a child Node}
\item{Missing}{Index of a row containing a child Node where are proceeded all observations with no value of the dividing feature}
\item{Prediction}{For leaves: Value of prediction in the leaf. For internal nodes: NA}
\item{Cover}{Number of observations seen by the internal node or collected by the leaf for the reference dataset}

\strong{data} - Dataset used as a reference for calculating SHAP values. A dataset passed to the \code{*.unify}, \code{unify} or \code{\link{set_reference_dataset}} function with \code{data} argument. A \code{data.frame}.

Object has two also attributes set:
\item{\code{model}}{A string. By what package the model was produced.}
\item{\code{missing_support}}{A boolean. Whether the model allows missing values to be present in explained dataset.}
}
\description{
\code{model_unified} object produced by \code{*.unify} or \code{unify} function.
}
\seealso{
\code{\link{unify}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_ranger_surv.R
\name{ranger_surv.unify}
\alias{ranger_surv.unify}
\title{Unify ranger survival model}
\usage{
ranger_surv.unify(
  rf_model,
  data,
  type = c("risk", "survival", "chf"),
  times = NULL
)
}
\arguments{
\item{rf_model}{An object of \code{ranger} class. At the moment, models built on data with categorical features
are not supported - please encode them before training.}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}

\item{type}{A character to define the type of model prediction to use. Either \code{"risk"} (default), which uses the risk score calculated as a sum of cumulative hazard function values, \code{"survival"}, which uses the survival probability at certain time-points for each observation, or \code{"chf"}, which used the cumulative hazard values at certain time-points for each observation.}

\item{times}{A numeric vector of unique death times at which the prediction should be evaluated. By default \code{unique.death.times} from model are used.}
}
\value{
For \code{type = "risk"} a unified model representation is returned - a \code{\link{model_unified.object}} object. For \code{type = "survival"} or \code{type = "chf"} - a \code{\link{model_unified_multioutput.object}} object is returned, which is a list that contains unified model representation (\code{\link{model_unified.object}} object) for each time point. In this case, the list names are time points at which the survival function was evaluated.
}
\description{
Convert your ranger model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\details{
The survival forest implemented in the \code{ranger} package stores cumulative hazard
functions (CHFs) in the leaves of survival trees, as proposed for Random Survival Forests
(Ishwaran et al. 2008). The final model prediction is made by averaging these CHFs
from all the trees. To provide explanations in the form of a survival function,
the CHFs from the leaves are converted into survival functions (SFs) using
the formula SF(t) = exp(-CHF(t)).
However, it is important to note that averaging these SFs does not yield the correct
model prediction as the model prediction is the average of CHFs transformed in the same way.
Therefore, when you obtain explanations based on the survival function,
they are only proxies and may not be fully consistent with the model predictions
obtained using for example \code{predict} function.
}
\examples{
if (requireNamespace("ranger", quietly = TRUE) &&
 requireNamespace("survival", quietly = TRUE)) {

  library(ranger)
  data_colon <- data.table::data.table(survival::colon)
  data_colon <- na.omit(data_colon[get("etype") == 2, ])
  surv_cols <- c("status", "time", "rx")
 
  feature_cols <- colnames(data_colon)[3:(ncol(data_colon) - 1)]
 
  train_x <- model.matrix(
    ~ -1 + .,
    data_colon[, .SD, .SDcols = setdiff(feature_cols, surv_cols[1:2])]
  )
  train_y <- survival::Surv(
    event = (data_colon[, get("status")] |>
               as.character() |>
               as.integer()),
    time = data_colon[, get("time")],
    type = "right"
  )
 
  rf <- ranger::ranger(
    x = train_x,
    y = train_y,
    data = data_colon,
    max.depth = 10,
    num.trees = 10
  )
  unified_model_risk <- ranger_surv.unify(rf, train_x, type = "risk")
  shaps <- treeshap(unified_model_risk, train_x[1:2,])
 
  # compute shaps for 3 selected time points
  unified_model_surv <- ranger_surv.unify(rf, train_x, type = "survival", times = c(23  , 50, 73))
  shaps_surv <- treeshap(unified_model_surv, train_x[1:2,])
}

}
\seealso{
\code{\link{ranger.unify}} for regression and classification \code{\link[ranger:ranger]{ranger models}}

\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fifa20.R
\docType{data}
\name{fifa20}
\alias{fifa20}
\title{Attributes of all players in FIFA 20}
\format{
A data frame with 18278 rows and 56 columns.
Most of variables representing skills are in range from 0 to 100 and will not be described here.
To list non obvious features:
\describe{
\item{overall}{Overall score of player's skills}
\item{potential}{Potential of a player, younger players tend to have higher level of potential}
\item{value_eur}{Market value of a player (in mln EUR)}
\item{international_reputation}{Range 1 to 5}
\item{weak_foot}{Range 1 to 5}
\item{skill_moves}{Range 1 to 5}
\item{work_rate}{Divided by slash levels of willingness to work in offense and defense respectively}
}
}
\source{
"Data has been scraped from the publicly available website \url{https://sofifa.com}"
\url{https://www.kaggle.com/stefanoleone992/fifa-20-complete-player-dataset}
}
\usage{
fifa20
}
\description{
Dataset consists of 56 columns, 55 numeric and one of type factor \code{'work_rate'}.
\code{value_eur} is a potential target feature.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_drwhy.R
\name{theme_drwhy}
\alias{theme_drwhy}
\alias{theme_drwhy_vertical}
\title{DrWhy Theme for ggplot objects}
\usage{
theme_drwhy()

theme_drwhy_vertical()
}
\value{
theme for ggplot2 objects
}
\description{
DrWhy Theme for ggplot objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify.R
\name{unify}
\alias{unify}
\title{Unify tree-based model}
\usage{
unify(model, data, ...)
}
\arguments{
\item{model}{A tree-based model object of any supported class (\code{gbm}, \code{lgb.Booster}, \code{randomForest}, \code{ranger}, or \code{xgb.Booster}).}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}

\item{...}{Additional parameters passed to the model-specific unification functions.}
}
\value{
A unified model representation - a \code{\link{model_unified.object}} object (for single-output models) or \code{\link{model_unified_multioutput.object}}, which is a list of \code{\link{model_unified.object}} objects (for multi-output models).
}
\description{
Convert your tree-based model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\examples{
if (
  requireNamespace("ranger", quietly = TRUE) &&
    requireNamespace("randomForest", quietly = TRUE)
) {
  library(ranger)
  data_fifa <- fifa20$data[
    !colnames(fifa20$data) \%in\%
      c(
        'work_rate',
        'value_eur',
        'gk_diving',
        'gk_handling',
        'gk_kicking',
        'gk_reflexes',
        'gk_speed',
        'gk_positioning'
      )
  ]
  data <- na.omit(cbind(data_fifa, target = fifa20$target)) 

  rf1 <- ranger::ranger(target ~ ., data = data, max.depth = 10, num.trees = 10)
  unified_model1 <- unify(rf1, data)
  shaps1 <- treeshap(unified_model1, data[1:2, ])
  plot_contribution(shaps1, obs = 1) 

  rf2 <- randomForest::randomForest(
    target ~ .,
    data = data,
    maxnodes = 10,
    ntree = 10
  )
  unified_model2 <- unify(rf2, data)
  shaps2 <- treeshap(unified_model2, data[1:2, ])
  plot_contribution(shaps2, obs = 1)
}

}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gpboost.unify}} for \code{\link[gpboost:gpboost]{GPBoost models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_gbm.R
\name{gbm.unify}
\alias{gbm.unify}
\title{Unify GBM model}
\usage{
gbm.unify(gbm_model, data)
}
\arguments{
\item{gbm_model}{An object of \code{gbm} class. At the moment, models built on data with categorical features
are not supported - please encode them before training.}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}
}
\value{
a unified model representation - a \code{\link{model_unified.object}} object
}
\description{
Convert your GBM model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\examples{
\donttest{
if (requireNamespace("gbm", quietly = TRUE)) {
  library(gbm)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  data['value_eur'] <- fifa20$target
  gbm_model <- gbm::gbm(
               formula = value_eur ~ .,
               data = data,
               distribution = "gaussian",
               n.trees = 20,
               interaction.depth = 4,
               n.cores = 1)
  unified_model <- gbm.unify(gbm_model, data)
  shaps <- treeshap(unified_model, data[1:2,])
  plot_contribution(shaps, obs = 1)
}}
}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_feature_dependence.R
\name{plot_feature_dependence}
\alias{plot_feature_dependence}
\title{SHAP value based Feature Dependence plot}
\usage{
plot_feature_dependence(
  treeshap,
  variable,
  title = "Feature Dependence",
  subtitle = NULL
)
}
\arguments{
\item{treeshap}{A treeshap object produced with the \code{\link{treeshap}} function. \code{\link{treeshap.object}}.}

\item{variable}{name or index of variable for which feature dependence will be plotted.}

\item{title}{the plot's title, by default \code{'Feature Dependence'}.}

\item{subtitle}{the plot's subtitle. By default no subtitle.}
}
\value{
a \code{ggplot2} object
}
\description{
Depending on the value of a variable: how does it contribute into the prediction?
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE) &&
 requireNamespace("scales", quietly = TRUE)) {
  library(xgboost)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target
  xgb_model <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 3,
   nrounds = 20,
   nthreads = 1
  )
  unified_model <- xgboost.unify(xgb_model, as.matrix(data))
  x <- head(data, 100)
  shaps <- treeshap(unified_model, x)
  plot_feature_dependence(shaps, variable = "overall")
}}
}
\seealso{
\code{\link{treeshap}} for calculation of SHAP values

\code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_interaction}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.model_unified}
\alias{predict.model_unified}
\title{Predict}
\usage{
\method{predict}{model_unified}(object, x, ...)
}
\arguments{
\item{object}{Unified model representation of the model created with a (model).unify function. \code{\link{model_unified.object}}}

\item{x}{Observations to predict. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model.}

\item{...}{other parameters}
}
\value{
a vector of predictions.
}
\description{
Predict using unified_model representation.
}
\examples{
\donttest{
if (requireNamespace("gbm", quietly = TRUE)) {
  library(gbm)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  data['value_eur'] <- fifa20$target
  gbm_model <- gbm::gbm(
    formula = value_eur ~ .,
    data = data,
    distribution = "laplace",
    n.trees = 20,
    interaction.depth = 4,
    n.cores = 1
  )
  unified <- gbm.unify(gbm_model, data)
  predict(unified, data[2001:2005, ])
}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{print.treeshap_multioutput}
\alias{print.treeshap_multioutput}
\title{Prints treeshap_multioutput objects}
\usage{
\method{print}{treeshap_multioutput}(x, ...)
}
\arguments{
\item{x}{a treeshap_multioutput object}

\item{...}{other arguments}
}
\value{
No return value, called for printing
}
\description{
Prints treeshap_multioutput objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_unified.R
\name{model_unified_multioutput.object}
\alias{model_unified_multioutput.object}
\title{Unified model representations for multi-output model}
\value{
List consisting of \code{model_unified} objects, one for each individual output of a model. For survival models, the list is named using the time points, for which predictions are calculated.
}
\description{
\code{model_unified_multioutput} object produced by \code{*.unify} or \code{unify} function.
}
\seealso{
\code{\link{unify}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_xgboost.R
\name{xgboost.unify}
\alias{xgboost.unify}
\title{Unify XGBoost model}
\usage{
xgboost.unify(xgb_model, data, recalculate = FALSE)
}
\arguments{
\item{xgb_model}{A XGBoost model - object of class \code{xgb.Booster}}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}

\item{recalculate}{logical indicating if covers should be recalculated according to the dataset given in data. Keep it \code{FALSE} if training data are used.}
}
\value{
a unified model representation - a \code{\link{model_unified.object}} object
}
\description{
Convert your XGBoost model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE)) {
  library(xgboost)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target
  xgb_model <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 3,
   nrounds = 20,
   nthreads = 1
  )
  unified_model <- xgboost.unify(xgb_model, as.matrix(data))
  shaps <- treeshap(unified_model, data[1:2,])
  plot_contribution(shaps, obs = 1)
}}

}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_gpboost.R
\name{gpboost.unify}
\alias{gpboost.unify}
\title{Unify GPBoost model}
\usage{
gpboost.unify(gpb_model, data, recalculate = FALSE)
}
\arguments{
\item{gpb_model}{A gpboost model - object of class \code{gpb.Booster}}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}

\item{recalculate}{logical indicating if covers should be recalculated according to the dataset given in data. Keep it \code{FALSE} if training data are used.}
}
\value{
a unified model representation - a \code{\link{model_unified.object}} object
}
\description{
Convert your GPBoost model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\examples{
\donttest{
if (requireNamespace("gpboost", quietly = TRUE)) {
  library(gpboost)
  param_gpb <- list(objective = "regression", max_depth = 2,
                     force_row_wise = TRUE, num_iterations = 20)
  data_fifa <- fifa20$data[!colnames(fifa20$data) \%in\%
               c('work_rate', 'value_eur', 'gk_diving', 'gk_handling',
               'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
  data <- na.omit(cbind(data_fifa, fifa20$target))
  sparse_data <- as.matrix(data[,-ncol(data)])
  x <- gpboost::gpb.Dataset(sparse_data, label = as.matrix(data[,ncol(data)]))
  gpb_data <- gpboost::gpb.Dataset.construct(x)
  gpb_model <- gpboost::gpboost(data = gpb_data, params = param_gpb,
                                  verbose = -1, num_threads = 1)
  unified_model <- gpboost.unify(gpb_model, sparse_data)
  shaps <- treeshap(unified_model, data[1:2, ])
  plot_contribution(shaps, obs = 1)
}}
}
\seealso{
\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_ranger.R
\name{ranger.unify}
\alias{ranger.unify}
\title{Unify ranger model}
\usage{
ranger.unify(rf_model, data)
}
\arguments{
\item{rf_model}{An object of \code{ranger} class. At the moment, models built on data with categorical features
are not supported - please encode them before training.}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}
}
\value{
a unified model representation - a \code{\link{model_unified.object}} object
}
\description{
Convert your ranger model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\examples{
if (requireNamespace("ranger", quietly = TRUE)) {
 library(ranger)
 data_fifa <- fifa20$data[!colnames(fifa20$data) \%in\%
                            c('work_rate', 'value_eur', 'gk_diving', 'gk_handling',
                             'gk_kicking', 'gk_reflexes', 'gk_speed', 'gk_positioning')]
 data <- na.omit(cbind(data_fifa, target = fifa20$target))

 rf <- ranger::ranger(target~., data = data, max.depth = 10, num.trees = 10)
 unified_model <- ranger.unify(rf, data)
 shaps <- treeshap(unified_model, data[1:2,])
 plot_contribution(shaps, obs = 1)
}
}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{treeshap_multioutput.object}
\alias{treeshap_multioutput.object}
\title{treeshap results for multi-output model}
\value{
List consisting of \code{treeshap} objects, one for each individual output of a model. For survival models, the list is named using the time points, for which TreeSHAP values are calculated.
}
\description{
\code{treeshap_multioutput} object produced by \code{treeshap} function.
}
\seealso{
\code{\link{treeshap}},

\code{\link{treeshap.object}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_reference_dataset.R
\name{set_reference_dataset}
\alias{set_reference_dataset}
\title{Set reference dataset}
\usage{
set_reference_dataset(unified_model, x)
}
\arguments{
\item{unified_model}{Unified model representation of the model created with a (model).unify function. (\code{\link{model_unified.object}}).}

\item{x}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model.}
}
\value{
\code{\link{model_unified.object}}. Unified representation of the model as created with a (model).unify function,
but with changed reference dataset (Cover column containing updated values).
}
\description{
Change a dataset used as reference for calculating SHAP values.
Reference dataset is initially set with \code{data} argument in unifying function.
Usually reference dataset is dataset used to train the model.
Important property of reference dataset is that SHAP values for each observation add up to its deviation from mean prediction for a reference dataset.
}
\examples{
\donttest{
if (requireNamespace("gbm", quietly = TRUE)) {
  library(gbm)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  data['value_eur'] <- fifa20$target
  gbm_model <- gbm::gbm(
  formula = value_eur ~ .,
    data = data,
    distribution = "laplace",
    n.trees = 20,
    interaction.depth = 4,
    n.cores = 1)
  unified <- gbm.unify(gbm_model, data)
  set_reference_dataset(unified, data[200:700, ])
}}
}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}

\code{\link{randomForest.unify}} for \code{\link[randomForest:randomForest]{randomForest models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{is.treeshap}
\alias{is.treeshap}
\title{Check whether object is a valid treeshap object}
\usage{
is.treeshap(x)
}
\arguments{
\item{x}{an object to check}
}
\value{
boolean
}
\description{
Does not check correctness of result, only basic checks
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_interaction.R
\name{plot_interaction}
\alias{plot_interaction}
\title{SHAP Interaction value plot}
\usage{
plot_interaction(
  treeshap,
  var1,
  var2,
  title = "SHAP Interaction Value Plot",
  subtitle = ""
)
}
\arguments{
\item{treeshap}{A treeshap object produced with \code{\link{treeshap}(interactions = TRUE)} function. \code{\link{treeshap.object}}.}

\item{var1}{name or index of the first variable - plotted on x axis.}

\item{var2}{name or index of the second variable - marked with color.}

\item{title}{the plot's title, by default \code{'SHAP Interaction Value Plot'}.}

\item{subtitle}{the plot's subtitle. By default no subtitle.}
}
\value{
a \code{ggplot2} object
}
\description{
This function plots SHAP Interaction value for two variables depending on the value of the first variable.
Value of the second variable is marked with the color.
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE) &&
 requireNamespace("scales", quietly = TRUE)) {
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target
  xgb_model2 <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 5,
   nrounds = 10,
   nthreads = 1
  )
  unified_model2 <- xgboost.unify(xgb_model2, data)
  inters <- treeshap(unified_model2, as.matrix(data[1:50, ]), interactions = TRUE)
  plot_interaction(inters, "dribbling", "defending")
}}
}
\seealso{
\code{\link{treeshap}} for calculation of SHAP Interaction values

\code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{treeshap}
\alias{treeshap}
\title{Calculate SHAP values of a tree ensemble model.}
\usage{
treeshap(unified_model, x, interactions = FALSE, verbose = TRUE)
}
\arguments{
\item{unified_model}{Unified data.frame representation of the model created with a (model).unify function. A \code{\link{model_unified.object}} object.}

\item{x}{Observations to be explained. A \code{data.frame} or \code{matrix} object with the same columns as in the training set of the model. Keep in mind that objects different than \code{data.frame} or plain \code{matrix} will cause an error or unpredictable behavior.}

\item{interactions}{Whether to calculate SHAP interaction values. By default is \code{FALSE}. Basic SHAP values are always calculated.}

\item{verbose}{Whether to print progress bar to the console. Should be logical. Progress bar will not be displayed on Windows.}
}
\value{
A \code{\link{treeshap.object}} object (for single-output models) or \code{\link{treeshap_multioutput.object}}, which is a list of \code{\link{treeshap.object}} objects (for multi-output models). SHAP values can be accessed from \code{\link{treeshap.object}} with \code{$shaps}, and interaction values can be accessed with \code{$interactions}.
}
\description{
Calculate SHAP values and optionally SHAP Interaction values.
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE)) {
  library(xgboost)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target

  # calculating simple SHAP values
  xgb_model <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 3,
   nrounds = 20,
   nthreads = 1
  )
  unified_model <- xgboost.unify(xgb_model, as.matrix(data))
  treeshap1 <- treeshap(unified_model, head(data, 3))
  plot_contribution(treeshap1, obs = 1)
  treeshap1$shaps

  # It's possible to calcualte explanation over different part of the data set

  unified_model_rec <- set_reference_dataset(unified_model, data[1:1000, ])
  treeshap_rec <- treeshap(unified_model, head(data, 3))
  plot_contribution(treeshap_rec, obs = 1)

  # calculating SHAP interaction values
  xgb_model2 <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 7,
   nrounds = 10,
   nthreads = 1
  )
  unified_model2 <- xgboost.unify(xgb_model2, as.matrix(data))
  treeshap2 <- treeshap(unified_model2, head(data, 3), interactions = TRUE)
  treeshap2$interactions
}}

}
\seealso{
\code{\link{xgboost.unify}} for \code{XGBoost models}
\code{\link{lightgbm.unify}} for \code{LightGBM models}
\code{\link{gpboost.unify}} for \code{GPBoost models}
\code{\link{gbm.unify}} for \code{GBM models}
\code{\link{randomForest.unify}} for \code{randomForest models}
\code{\link{ranger.unify}} for \code{ranger models}
\code{\link{ranger_surv.unify}} for \code{ranger survival models}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_drwhy.R
\name{colors_discrete_drwhy}
\alias{colors_discrete_drwhy}
\alias{colors_breakdown_drwhy}
\title{DrWhy color palettes for ggplot objects}
\usage{
colors_discrete_drwhy(n = 2)

colors_breakdown_drwhy()
}
\arguments{
\item{n}{number of colors for color palette}
}
\value{
color palette as vector of characters
}
\description{
DrWhy color palettes for ggplot objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeshap.R
\name{treeshap.object}
\alias{treeshap.object}
\title{treeshap results}
\value{
List consisting of four elements:
\describe{
\item{shaps}{A \code{data.frame} with M columns, X rows (M - number of features, X - number of explained observations). Every row corresponds to SHAP values for a observation. }
\item{interactions}{An \code{array} with dimensions (M, M, X) (M - number of features, X - number of explained observations). Every \code{[, , i]} slice is a symmetric matrix - SHAP Interaction values for a observation. \code{[a, b, i]} element is SHAP Interaction value of features \code{a} and \code{b} for observation \code{i}. Is \code{NULL} if interactions where not calculated (parameter \code{interactions} set \code{FALSE}.) }
\item{unified_model}{An object of type \code{\link{model_unified.object}}. Unified representation of a model for which SHAP values were calculated. It is used by some of the plotting functions.}
\item{observations}{Explained dataset. \code{data.frame} or \code{matrix}. It is used by some of the plotting functions.}
}
}
\description{
\code{treeshap} object produced by \code{treeshap} function.
}
\seealso{
\code{\link{treeshap}},

\code{\link{plot_contribution}}, \code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}, \code{\link{plot_interaction}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unify_randomForest.R
\name{randomForest.unify}
\alias{randomForest.unify}
\title{Unify randomForest model}
\usage{
randomForest.unify(rf_model, data)
}
\arguments{
\item{rf_model}{An object of \code{randomForest} class. At the moment, models built on data with categorical features
are not supported - please encode them before training.}

\item{data}{Reference dataset. A \code{data.frame} or \code{matrix} with the same columns as in the training set of the model. Usually dataset used to train model.}
}
\value{
a unified model representation - a \code{\link{model_unified.object}} object
}
\description{
Convert your randomForest model into a standardized representation.
The returned representation is easy to be interpreted by the user and ready to be used as an argument in \code{treeshap()} function.
}
\details{
Binary classification models with a target variable that is a factor with two levels, 0 and 1, are supported
}
\examples{
if (requireNamespace("randomForest", quietly = TRUE)) {
  library(randomForest)
  data_fifa <- fifa20$data[
    !colnames(fifa20$data) \%in\%
      c(
        'work_rate',
        'value_eur',
        'gk_diving',
        'gk_handling',
        'gk_kicking',
        'gk_reflexes',
        'gk_speed',
        'gk_positioning'
      )
  ]
  data <- na.omit(cbind(data_fifa, target = fifa20$target))

  rf <- randomForest::randomForest(
    target ~ .,
    data = data,
    maxnodes = 10,
    ntree = 10
  )
  unified_model <- randomForest.unify(rf, data)
  shaps <- treeshap(unified_model, data[1:2, ])
   # plot_contribution(shaps, obs = 1)
}

}
\seealso{
\code{\link{lightgbm.unify}} for \code{\link[lightgbm:lightgbm]{LightGBM models}}

\code{\link{gbm.unify}} for \code{\link[gbm:gbm]{GBM models}}

\code{\link{xgboost.unify}} for \code{\link[xgboost:xgboost]{XGBoost models}}

\code{\link{ranger.unify}} for \code{\link[ranger:ranger]{ranger models}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_contribution.R
\name{plot_contribution}
\alias{plot_contribution}
\title{SHAP value based Break-Down plot}
\usage{
plot_contribution(
  treeshap,
  obs = 1,
  max_vars = 5,
  min_max = NA,
  digits = 3,
  explain_deviation = FALSE,
  title = "SHAP Break-Down",
  subtitle = ""
)
}
\arguments{
\item{treeshap}{A treeshap object produced with the \code{\link{treeshap}} function. \code{\link{treeshap.object}}.}

\item{obs}{A numeric indicating which observation should be plotted. Be default it's first observation.}

\item{max_vars}{maximum number of variables that shall be presented. Variables with the highest importance will be presented.
Remaining variables will be summed into one additional contribution. By default \code{5}.}

\item{min_max}{a range of OX axis. By default \code{NA}, therefore it will be extracted from the contributions of \code{x}.
But it can be set to some constants, useful if these plots are to be used for comparisons.}

\item{digits}{number of decimal places (\code{\link{round}}) to be used.}

\item{explain_deviation}{if \code{TRUE} then instead of explaining prediction and plotting intercept bar, only deviation from mean prediction of the reference dataset will be explained. By default \code{FALSE}.}

\item{title}{the plot's title, by default \code{'SHAP Break-Down'}.}

\item{subtitle}{the plot's subtitle. By default no subtitle.}
}
\value{
a \code{ggplot2} object
}
\description{
This function plots contributions of features into the prediction for a single observation.
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE) &&
 requireNamespace("scales", quietly = TRUE)) {
  library(xgboost)
  data <- fifa20$data[colnames(fifa20$data) != 'work_rate']
  target <- fifa20$target
  xgb_model <- xgboost::xgboost(
   x = as.matrix(data),
   y = target,
   objective = "reg:squarederror",
   max_depth = 3,
   nrounds = 20,
   nthreads = 1
  )
  unified_model <- xgboost.unify(xgb_model, as.matrix(data))
  x <- head(data, 1)
  shap <- treeshap(unified_model, x)
  plot_contribution(shap, 1, min_max = c(-5e5, 1.3e8))
}}

}
\seealso{
\code{\link{treeshap}} for calculation of SHAP values

\code{\link{plot_feature_importance}}, \code{\link{plot_feature_dependence}}, \code{\link{plot_interaction}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_unified.R
\name{print.model_unified}
\alias{print.model_unified}
\title{Prints model_unified objects}
\usage{
\method{print}{model_unified}(x, ...)
}
\arguments{
\item{x}{a model_unified object}

\item{...}{other arguments}
}
\value{
No return value, called for printing
}
\description{
Prints model_unified objects
}
